//          AI  21-Feb-08 Read file with relevant information for the transfer
// CR x1    jcv  7-Oct-08 Changed for getting raw files from local or remote
// SPR x7   jcv 22-Jan-09 Processing may be exigent or not about rawtm quality
// CR x11   jcv 31-Mar-09 Fixed tracing verbosity 
// CR x12   jcv  1-Apr-09 Refactored the DU Processing
// SPR x11  jcv  2-Apr-09 Added checking information to DU Processor 
// SPR x12  jcv 17-Apr-09 Removed usage of lib/data/rawtmdata/info.dat  
// CRx14    jcv 17-Apr-09 Infrastructure for tmproc dumper/fixer 
// SPRx14   jcv 21-Apr-09 Enhanced dump funcionality, based on strict run
// CRx15    jcv           tmproc able to filter by APID-Type-Subtype
// SPRx16   jcv 19-May-09 Time in DFH guessing capabilities (CH heritage!) 
// SPRx17   jcv  8-Jun-09 Additional DU removal based in DU type
// CRx16    jcv 11-Jun-09 Statistics on DU type 
// SPRx32   jcv 1-Jul-10  Enhanced verbosity for gaps stamps
// CRx20  jcv 21-Sep-10 BTI with GS ids to ignore

#include <iostream>
#include <utility>
#include <iterator>
#include <set>
#include <fstream>

#include "errstr.h"
#include "param.h"

#include "ProcessorList.h"
#include "DataDefinition.h"
#include "DUExtractor.h"
#include "DUProcessor.h"
#include "Utils.h"
#include "BTI.h"
#include "SasFileLocator.h"	// AI: 2008_02_21: for locating infoFile 

// CRx15
#include "FilterDefinition.h"
// END CRx15

// SPRx16 Time in DFH guessing capabilities
#include "dfhtimestamp.h"
// END SPRx16

// NEw SPRx14 filter file management
void readFilterFile(string sFilename, FilterInfo * AppliedFilters);
// END SPRx14

// NEW SPRx16 DHF guessing
void vtmzz_fiReadNoDFHTimestampFile( int *NofGeneric, 
                                     int *NofSpecific,
                                     VTMZZ_TNODFHTIME *MyarNoDFHTime);
void fvSelectionSort(
         VTMZZ_TNODFHTIME * aiData, int iNumEntries );
// END SPRx16

DUExtractor::DUExtractor(STime start, STime stop) {
  startTime = start;
  stopTime  = stop;

  string srcurl;

// New JCV SPRx12
  srcurl = "ftp://user:pass@machine/";

  tmArchive = new DataFiles(srcurl,
    stringParameter("cachedir"), booleanParameter("useS2K"));

  hasTimeFilter = false;

  // CRx15 Additional initialisation for reading filter information
  string sFilterFile;

  AppliedFilterInfo.bFilterRequired=false;
  AppliedFilterInfo.bRemoveRequired=false;

  // NEW CRx20
  bool bFullBTI_=booleanParameter("fullBTI");
  if (bFullBTI_) 
  errstr << "GSid info in BTI file IS used " 
            << Sas::message(Msg::App, Msg::Sparse); 
  else
  errstr << "GSid info in BTI file is NOT used " 
            << Sas::message(Msg::App, Msg::Sparse); 
  // END CRx20


  if(std::getenv("RWUT_FILTERFILE") == 0) {
    errstr << "RWUT_FILTERFILE is not defined " << Sas::message(Msg::App, Msg::Sparse); 
  } else {
    errstr << "RWUT_FILTERFILE defined " << Sas::message(Msg::App, Msg::Sparse); 
    sFilterFile = std::getenv("RWUT_FILTERFILE");
    readFilterFile(sFilterFile,&AppliedFilterInfo);

  } // end else file required
  
  // END CRx15

  // NEw SPRx16 Read Time in DFH config file
  iNumGeneric    = 0;
  iNumSpecific   = 0;
  vtmzz_fiReadNoDFHTimestampFile(&iNumGeneric,&iNumSpecific,arNoDFHTime);  
  // END SPRx16

}

// NEW SPRx14 Filter File management
void readFilterFile(string sFilename, FilterInfo * AppliedFilters) {

  // Filter Fields /////////////
  string sDummie;
  string sDirective;

  // Fields for DIRECTIVE-A, APID-Type-... filtering
  string sFilterApid;
  string sFilterType;
  string sFilterSubType;
  string sFilterSID;
  string sMask;
  string sPacketFileName;

  // Fields for DIRECTIVE-B, DU removal  based in Time and Type 
  string sRemovalStart;
  string sRemovalEnd;
  // SPRx17 Removal based in DU Type
  string sKeepDU;


  string sLocatedFile = fileLocator.locate(sFilename, "", false, true);
  if (sLocatedFile.length() == 0)
    {
      errstr << "But file seems not to exist. " << Sas::error("NotFilterFile");
      exit(1);
    }

  ifstream ifs(sLocatedFile.c_str());
  if (! ifs.good()) {
         errstr << "file stream invalid. " << Sas::error("BadStream");
         exit(1);
  }

  try {
        getline(ifs, sDummie); 
        getline(ifs, sDirective); 

        if (sDirective.find("DIRECTIVE-A-") != string::npos) {
              // DIRECTIVE A, Packet Details filtering
              errstr << "DIRECTIVE A, Packet Details filtering" 
                     << Sas::message(Msg::App, Msg::Verbose); 

              getline(ifs, sFilterApid); 
              sFilterApid=sFilterApid.substr(10);
              getline(ifs, sFilterType); 
              sFilterType=sFilterType.substr(10);
              getline(ifs, sFilterSubType); 
              sFilterSubType=sFilterSubType.substr(10);
              getline(ifs, sFilterSID); 
              sFilterSID=sFilterSID.substr(10);
              getline(ifs, sMask); 
              sMask=sMask.substr(10);

              getline(ifs, sPacketFileName); 
              sPacketFileName=sPacketFileName.substr(10);

              AppliedFilters->bFilterRequired=true;
              AppliedFilters->FilterAPID=atoi(sFilterApid.c_str());
              AppliedFilters->FilterType=atoi(sFilterType.c_str());
              AppliedFilters->FilterSubType=atoi(sFilterSubType.c_str());
              AppliedFilters->FilterSID=atoi(sFilterSID.c_str());
              AppliedFilters->FilterMask=sMask;

              AppliedFilters->ofsPacketFile.open(sPacketFileName.c_str());
              if (!AppliedFilters->ofsPacketFile.good()) {
                errstr << "Failed to open packet dump file " << sPacketFileName 
                    << Sas::fatal("failOpenDumpfile");
              } 

        } else {
             // DIRECTIVE B, Start-End Time DU removal 
             errstr << "DIRECTIVE B, Start-End Time DU removal" 
                     << Sas::message(Msg::App, Msg::Verbose); 
              getline(ifs, sRemovalStart); 
              sRemovalStart=sRemovalStart.substr(10);
              getline(ifs, sRemovalEnd); 
              sRemovalEnd=sRemovalEnd.substr(10);
              // NEW SPRx17
              getline(ifs, sKeepDU); 
              sKeepDU=sKeepDU.substr(10);

              AppliedFilters->bRemoveRequired=true;
              AppliedFilters->RemoveEnd=sRemovalEnd;
              AppliedFilters->RemoveStart=sRemovalStart;
              // NEW SPRx17 
              AppliedFilters->KeepDUTypes=sKeepDU;


        } // End directive branching 

  } catch(...) {  // The try failed
        errstr << "Syntax err in filter file." << Sas::error("SyntaxFile");
        exit(1);
  }

}


// NEW SPRx16
void fvSelectionSort(
         VTMZZ_TNODFHTIME * aiData,  int iNumEntries ) 
{

/*+DATA************************************/

   int      iLoop1;
   int      iLoop2;
   VTMZZ_TNODFHTIME      iTemp;
   int      iMin;

   //vector<long int> myvector(aiData,aiData+iNumEntries);
   //vector<long int>::iterator it;

/*-DATA************************************/
   
   for( iLoop1 = 0; iLoop1 < iNumEntries ; iLoop1++ )
    errstr << "Before sorting, [" << iLoop1 << "]" 
          << aiData[ iLoop1 ].uMap.iPktID 
          << Sas::message(Msg::App, Msg::Noisy);

   /*+ Sort the data into ascending order. +*/

   //sort(myvector.begin(),myvector.begin()+iNumEntries);

   for( iLoop1 = 0; iLoop1 < iNumEntries ; iLoop1++ )
   {
      iMin = iLoop1;

      for( iLoop2 = iLoop1+1 ; iLoop2 < iNumEntries; iLoop2++ )
         if( aiData[ iLoop2 ].uMap.iPktID < aiData[ iMin ].uMap.iPktID ) 
              iMin = iLoop2;

      iTemp            = aiData[ iMin ];
      aiData[ iMin ]   = aiData[ iLoop1 ];
      aiData[ iLoop1 ] = iTemp;
   }

   for( iLoop1 = 0; iLoop1 < iNumEntries ; iLoop1++ )
    errstr << "After sorting, [" << iLoop1 << "]" 
          << aiData[ iLoop1 ].uMap.iPktID 
          << Sas::message(Msg::App, Msg::Noisy);

} /* END fvSelectionSort */

void vtmzz_fiReadNoDFHTimestampFile( int *NofGeneric, 
                                     int *NofSpecific,
                                     VTMZZ_TNODFHTIME *MyarNoDFHTime)
{

/*+DATA************************************/

   int                   iNumFieldsRead;     /* Number of fields read */
   int                   iTotal;             /* Total number of lines read */
   char    acBuffer[ 2048 ];   /* One line of the file */
   short   siAPID;             /* Apid read from file */

   short   sType;              /* Pkt type read from file */
   short   sSubtype;           /* Pkt subtype read from file */
   /* File containing packets which do NOT include OBT */

   FILE                  *pfNoTime;         

   int                   iLoop;              /* Loop counter */
   int                   iIndex;              /* Loop counter */

/*-DATA************************************/

   /*+ Initilise variables. +*/
   string sNoDFHFile;
   int iMyNofGeneric  = 0;
   int iMyNofSpecific = 0;
   iTotal         = -1;

   /*+ Open file containing packet information as read only +*/
   /*+ IF we cannot open the file +*/
      /*+ set return status and exit +*/
   /*+ END IF +*/

   if(std::getenv("VTMZZ_NO_DFH_TIME_LOGICAL") == 0) 
      errstr << "VTMZZ_NO_DFH_TIME_LOGICAL not defined " << Sas::error("NoDFHFileDef");

   sNoDFHFile = std::getenv("VTMZZ_NO_DFH_TIME_LOGICAL");
   if( (pfNoTime = fopen( sNoDFHFile.c_str(), "r" )) == NULL )
   {
      errstr << "Cannot open the file " <<  VTMZZ_NO_DFH_TIME_LOGICAL << Sas::error("DFHFileErr");
      exit(1);
   }

   /*+ Read in records. Skip lines beginning with '!' +*/

   while ( fgets( acBuffer, (sizeof acBuffer), pfNoTime ) != NULL )
   {
      /* Read in records info. Note that records with a "!" as the first */
      /* character are treated as comment lines and are not further      */
      /* processed. All other records are processed to extract the       */
      /* required Type, Subtype and APID. Only those records where       */
      /* the Type, Subtype and APID lie in the following ranges are used */
      /* to determine the TM packets which do not have the time          */
      /* specified in the Data Field Header;                             */
      /*                                                                 */
      /* Type;      0 - 15,                                              */
      /* Subtype;   0 - 15,                                              */
      /* APID;     -1 - 2047                                             */
      /*                                                                 */
      /* Records which do not have Type, subtype and APID values         */
      /* in this range are ignored.                                      */
      /* NOTE: If -1 is specified as the APID this is taken to implie    */
      /*       that this is a generic specification, i.e. all packets    */
      /*       of the specified type and subtype do not have time field  */
      /*       in the data field header.                                 */

      errstr << "File: " << VTMZZ_NO_DFH_TIME_LOGICAL
             << " Reading Line: " << acBuffer 
            << Sas::message(Msg::App, Msg::Noisy);
   
      //errstr << "Logging short size " << sizeof(short) 
      //    << " int size " << sizeof(int)
      //    << " char size " << sizeof(char)
      //    << Sas::message(Msg::App, Msg::Verbose);

      if( acBuffer[ 0 ] != '!' )
      {
         iNumFieldsRead = sscanf( acBuffer, VTMZZ_NO_DFH_TIME_PKTS_FMT,
            &sType,
            &sSubtype,
            &siAPID );

         //errstr << "Entry in file sType:   " << sType 
         //     << " sSubType: " << sSubtype 
         //     << " siAPID: " << siAPID
         //     << Sas::message(Msg::App, Msg::Noisy);

         if( ( iNumFieldsRead == 3 ) &&
            (( sType >= 0 ) && ( sType <= 15 )) &&
            (( sSubtype >= 0 ) && ( sSubtype <= 15 )) &&
            (( siAPID >= -1 ) && ( siAPID <= 2047 )) )
         {
            /* if the number of entries is greater than the maximum number
               allowable, then set return status and exit. */

            if( ++iTotal > VTMZZ_MAX_NO_DFH_TIME_PACKETS )
            {
               
               errstr << "Too many entries, exceeded " <<  VTMZZ_MAX_NO_DFH_TIME_PACKETS << Sas::error("TooManyDFH");
               exit(1);
            }

            /* Valid record, check if generic specification. */

            if( siAPID == -1 )
            {
               /* Generic specifcation, insert into generic array. */
              errstr << iTotal << " is a Generic entry" 
                << Sas::message(Msg::App, Msg::Noisy);

              iMyNofGeneric++;
            }
            else
            {
               /* Insert into specific array. */
              errstr << iTotal << " is a Specific entry" 
                << Sas::message(Msg::App, Msg::Noisy);

              iMyNofSpecific++;
            }

            MyarNoDFHTime[ iTotal ].uMap.rPktID.cType    = sType;
            MyarNoDFHTime[ iTotal ].uMap.rPktID.cSubtype = sSubtype;
            MyarNoDFHTime[ iTotal ].uMap.rPktID.siAPID   = siAPID;

            errstr << "----PktID " << MyarNoDFHTime[ iTotal ].uMap.iPktID
                << "----cType " << hex << setw(2) << setfill('0')
                << (int)MyarNoDFHTime[ iTotal ].uMap.rPktID.cType
                << dec << " cSubType " << hex  << setw(2) << setfill('0') 
                << (int)MyarNoDFHTime[ iTotal ].uMap.rPktID.cSubtype
                << dec << " siAPID " 
                << MyarNoDFHTime[ iTotal ].uMap.rPktID.siAPID
                << Sas::message(Msg::App, Msg::Noisy);


         } // END of properly read line (3 args)
      } // END of comment or non-comment line
   } // End while the lines are being read

   errstr << "Read " << iTotal+1 << " entries , which are  " 
          << iMyNofGeneric << "generic, and are " 
          << iMyNofSpecific <<  " specific "   
          << Sas::message(Msg::App, Msg::Noisy);

   /*+ Sort the arrays into ascending order. +*/
   /* This will have the result that all generic entries will come before
      any specifc entries. */

   //fvSelectionSort( ( long int * ) MyarNoDFHTime, iTotal );
   fvSelectionSort( MyarNoDFHTime, iTotal+1 );

   /* Log those packets which do not have time stamps in the DFH.  */

   errstr << "After sorting, Generic Packets which do not have an OBT " << Sas::message(Msg::App, Msg::Noisy);

   for( iLoop = 0; iLoop < iMyNofGeneric; iLoop++)
   {
      /* Process (log) all generic entries. */
      errstr << "----[" << iLoop << "] PktID " 
                << MyarNoDFHTime[ iLoop ].uMap.iPktID
                << " Type: " 
                << (int)MyarNoDFHTime[ iLoop ].uMap.rPktID.cType
                << " SubType:  " 
                << (int)MyarNoDFHTime[ iLoop ].uMap.rPktID.cSubtype
                << Sas::message(Msg::App, Msg::Noisy);
   }

   errstr << "After sorting, logging Specific Packets which do not have an OBT " << Sas::message(Msg::App, Msg::Noisy);

   for( iLoop = 0;
        (iLoop < iMyNofSpecific );
        iLoop++)
   {
      /* Process (log) all Specific entries.  */
      iIndex=iLoop+iMyNofGeneric;
      errstr << "----[" << iIndex << "] PktID " 
                << MyarNoDFHTime[ iIndex  ].uMap.iPktID
                << " Type: " 
                << (int)MyarNoDFHTime[ iIndex ].uMap.rPktID.cType
                << " SubType:  " 
                << (int)MyarNoDFHTime[ iIndex ].uMap.rPktID.cSubtype
                << " APID: " <<  MyarNoDFHTime[ iIndex ].uMap.rPktID.siAPID 
                << Sas::message(Msg::App, Msg::Noisy);
   }

   /* Close the file. */

   fclose( pfNoTime );

   // Return success
   *NofGeneric=iMyNofGeneric;
   *NofSpecific=iMyNofSpecific;
 

} /* END vtmzz_fiReadNoDFHTimestampFile */      

// END SPRx16


void DUExtractor::registerProcessor(DUProcessor *proc) {

  Listeners[proc->processorType()].addProcessor(proc);

}

void DUExtractor::unregisterProcessor(DUProcessor *proc) {
	
  Listeners[proc->processorType()].removeProcessor(proc);

}

// NEW CR x12
void DUExtractor::fetchDU() {

  // Determine which data files are required and fetches from remote dir

  // The "remote" names are generated for given start-stop interval
  // (i.e. they will contain the ftp command, indeed)
  std::vector<string> archiveFiles = tmArchive->remoteDataFiles(startTime, stopTime);

  for (std::vector<string>::const_iterator iter = archiveFiles.begin();
       iter != archiveFiles.end();
       iter++) {

    // CR x1 starts
    //s << "wget -nv -nc -P" << tmArchive->getDirectory() << " -nH --cut-dirs=2 " << *iter;
    // rwut_getraw.sh <cache dir> <ftp command to use>
    stringstream s;
    s << "rwut_getraw.sh " << tmArchive->getDirectory() << " " << *iter;

    errstr << "Fetching data file " << s.str() << Sas::message(Msg::App, Msg::Sparse);

    system(s.str().c_str());

  }


}

// NEW CR x12
void DUExtractor::loopDU() {

  int maxDUsize;
  bool myloop = true;

  // Read the loop parameters


  // NEW SPR x7
  bool bStrictRun=booleanParameter("strict");
  if(  bStrictRun )
    errstr << "Strict=yes Running in normal mode" << Sas::warning("Strict_yes");
  else
    errstr << "Strict=no WARNING Running in non-abort mode" << Sas::warning("Strict_no");
  // END SPR x7

  // Begin the preparation of the loop

  errstr << "Data Limits: " << startTime << " - " << stopTime << Sas::message(Msg::App, Msg::Sparse);

  // Check that start < stop
  if(startTime >= stopTime) {
    errstr << "Start time (" << startTime << ") later than Stop time (" << stopTime << ")"
           << Sas::fatal("startTimeGTstopTime");
  }

  // Fetches the remote files
  this->fetchDU();

  // Once they are local, now loop on them locally

  std::vector<string> dataFiles = tmArchive->localDataFiles(startTime,stopTime);

  for (std::vector<string>::const_iterator iter = dataFiles.begin();
       iter != dataFiles.end() && myloop;
       iter++) {

    // Open the file
    errstr << "Opening data file " << *iter << Sas::message(Msg::App, Msg::Sparse);

    if ((currentFile = fopen((*iter).c_str(),"r")) == NULL) {
      errstr << "Failed to open data file " << *iter
             << Sas::fatal("failOpenDatafile");
    }

    this->processDU(currentFile,&maxDUsize,&myloop,bStrictRun);

    // Close data file
    errstr << "Closing data file " << *iter << Sas::message(Msg::App, Msg::Sparse);
    fclose(currentFile);


  } // END FOR loop files

  // What was the biggest DU found
  errstr << "Biggest DU found " << dec << maxDUsize << Sas::message(Msg::App, Msg::Sparse);


} 

// NEW CRx14 Infrastructure for dumper

void DUExtractor::dumpDU(FILE * rawFile, FILE * outFile, int *maxDU, bool bRunStrict ) {

 // TEST
 *maxDU=0;
 //bRunStrict = false;

 // CRx15
 if (AppliedFilterInfo.bFilterRequired)  
  errstr << "Detailed Dumps Filters Activated" << Sas::message(Msg::App, Msg::Verbose); 
 else
  errstr << "No Filters to be used for dumping packets" << Sas::message(Msg::App, Msg::Verbose); 

 if (AppliedFilterInfo.bRemoveRequired)  
  errstr << "Removal of DU Activated" << Sas::message(Msg::App, Msg::Verbose); 
 else
  errstr << "No removal for DU writing" << Sas::message(Msg::App, Msg::Verbose); 
 // END CRx15

  RawTM::XMCS::DataUnit *DU;

  unsigned char rawdata[RawTM::XMCS::max_size];
  
  STime saneMinTime = STime("1998-01-01T00:00:00.0");
  STime saneMaxTime = STime("2025-01-01T00:00:00.0");
  STime lastValidTime;

  STime lastTime;
  STime firstTime;
  //STime firstDUTime;

  // CRx16 Statistics on DU types
  int aDUCounter[8];
  int nFrames = 0;


  // SPRx14 
  bool badDU = false;
  bool writeDU = true;
  // END SPRx14

  //bool firstDU = true;
  //bool processDU = true;
  //int maxSize = 0;
  lastDUTime = STime::earliest;
  lastSequenceCount = 0;

  // List of timestamps received
  std::set<signed long long> timeList;

  hasTimeFilter = true;
  
  BTI bti;

  // Reset error count
  //int nErrors = 0;

  // CRx16 Statistics on DU types
  for (unsigned int i=0; i<8; i++) aDUCounter[i]=0; 

  ////////////////////////////////////////////////////////
  // Read the data: TMDU Frame Header
  while (fread((void *)rawdata, RawTM::XMCS::header_size, 1, rawFile)) {

  nFrames++;
  errstr << "--------------DU " << nFrames 
         << " READ -----------" << Sas::message(Msg::App, Msg::Verbose); 

  // SPRx14 
  badDU = false;
  writeDU=true;
  // END SPRx14

  char isS2K = rawdata[15];

  // Byte swap fields
  Utils::vmsByteSwap(&rawdata[0], 2); // Size

  if (isS2K == 0x1) { // SCOS-I == 0xAA, SCOS-2000 == 0x1 
          errstr << "Is SCOS-2000 file" << Sas::message(Msg::App, Msg::Verbose);
          Utils::sunByteSwap(&rawdata[2], 4); // Time
          Utils::sunByteSwap(&rawdata[6], 4); // Time
          Utils::sunByteSwap(&rawdata[10], 4); // Counter
          }
          else {
          errstr << "Is SCOS-I file" << Sas::message(Msg::App, Msg::Verbose);
          Utils::vmsByteSwap(&rawdata[2], 8); // Time
          Utils::vmsByteSwap(&rawdata[10], 4); // Counter
  }

  // Cast header onto header record
  DU = (RawTM::XMCS::DataUnit *)rawdata;

  // Checks 
  if (DU->Size < 17) {
   errstr << "WARNING Invalid DU Size found" << DU->Size  << Sas::warning("duInvalidSize");
   errstr << "WARNING Invalid DU Size found" << DU->Size  << Sas::message(Msg::App, Msg::Verbose);
  // SPRx14 
  badDU = true;
  // END SPRx14
  }

  if ((DU->Type < 1) || (DU->Type > 7)) {
   errstr << "WARNING Invalid DU Type found" << DU->Type << Sas::warning("duInvalidType");
   errstr << "WARNING Invalid DU Type found" << DU->Type << Sas::message(Msg::App, Msg::Verbose);
  // SPRx14 
  badDU = true;
  // END SPRx14
  }
  // CRx16 Statistics on DU Type
  else
     aDUCounter[DU->Type]++;

  // Get the DU time
  STime duTime;
  if (isS2K == 0x1) {     
         duTime = Utils::readTDSTime(DU->TDSTime);
      }
  else {
         duTime = Utils::readVMSTime(DU->VMSTime);
  }

  if ((duTime < saneMinTime) || (duTime > saneMaxTime)) {
   errstr << "WARNING Invalid DU Sane Time found" << Sas::warning("duInvalidSane");
   errstr << "WARNING Invalid DU Sane Time found" << Sas::message(Msg::App, Msg::Verbose);
  // SPRx14 
  badDU = true;
  // END SPRx14
  }

  // CRx16 Additional statistics
  if (nFrames == 1) firstTime=duTime;
  lastTime=duTime;
  // END CRx16


  // Verbose output  
  errstr << "DU of" 
             << " Type: " << (int)DU->Type 
             << " Size: " << dec << DU->Size 
             << " Header size " << RawTM::XMCS::header_size 
             << " Counter: " << hex << DU->Counter << dec
             << " Time: " << duTime.toString(7)
             << Sas::message(Msg::App, Msg::Verbose);

  ////////////////////////////////////////////////////////
  // Payload processing
  // Read the data: read the payload
  fread((void *)(rawdata + RawTM::XMCS::header_size), DU->Size - RawTM::XMCS::header_size , 1, rawFile);

  // Insert timestamp into list
  timeList.insert(DU->VMSTime);

  // NEw SPRx14
  // If the removal is activated, check and flag properly
  if (AppliedFilterInfo.bRemoveRequired)  
    {
    // SPRx17 Added removal by checking the DU type
    // If not found in the filter string, then do not keep it 
    bool bDUKeep = true;
    std::ostringstream ostrDUType;
    ostrDUType << (int)DU->Type;
    errstr << " Searching  DU type " << ostrDUType.str() 
           << " in filter string " << AppliedFilterInfo.KeepDUTypes
           << Sas::message(Msg::App, Msg::Noisy);
    if ( AppliedFilterInfo.KeepDUTypes.find(ostrDUType.str()) == string::npos) 
       {
         bDUKeep = false;
         errstr << " Not found in string" << Sas::message(Msg::App, Msg::Noisy);
       } else {
         errstr << " found in string" << Sas::message(Msg::App, Msg::Noisy);
       }
    if (  (duTime > AppliedFilterInfo.RemoveStart ) &&
          (duTime < AppliedFilterInfo.RemoveEnd )  &&
          (bDUKeep == false )  )
     {
       writeDU=false;
       errstr << "This DU will be REMOVED!!!" 
              << Sas::message(Msg::App, Msg::Verbose);
     }


    } // End If filters required

  // CRx16 Statistics here
  lastDUTime = duTime;

  // If the DU is earlier than we want, or later, skip to next
  if (  (duTime > startTime ) && (duTime < stopTime)   ) {
  // END SPRx14

  // Send DU to the processors
  for(unsigned int i = 0; i < Listeners[(DUProcessor::type)DU->Type].getMap().size(); i++) {
    DUProcessor *p = (DUProcessor *)Listeners[(DUProcessor::type)DU->Type].getMap()[i];
    // NEW CRx15 Applied Filters
    //p->procData(DU);
    // SPRx16 DFH capabilities
    //p->procData(DU,&AppliedFilterInfo);
    p->procData(DU,&AppliedFilterInfo,iNumGeneric,iNumSpecific,arNoDFHTime);
  }         

  // NEW SPRx14
  } // End of duTime within limits
  // END SPRx14

  /////////////////////////////////////////////////////////////////
  // Dump the TMDU
  // RAWDATA HAS BEEN MODIFIED
  /////////////////////////////////////////////////////////////////

  // Byte swap fields
  Utils::vmsByteSwap(&rawdata[0], 2); // Size

  if (isS2K == 0x1) { // SCOS-I == 0xAA, SCOS-2000 == 0x1 
      errstr << "Undoing SCOS-2000 " << Sas::message(Msg::App, Msg::Verbose);
      Utils::sunByteSwap(&rawdata[2], 4); // Time
      Utils::sunByteSwap(&rawdata[6], 4); // Time
      Utils::sunByteSwap(&rawdata[10], 4); // Counter
          }
          else {
      errstr << "Undoing SCOS-I " << Sas::message(Msg::App, Msg::Verbose);
      Utils::vmsByteSwap(&rawdata[2], 8); // Time
      Utils::vmsByteSwap(&rawdata[10], 4); // Counter
  }

  // End of swaps

  // SPRx14 It will only dump the DU if ,
  //  Strict = yes, or
  //  Strict but it is a good DU
  //  and writeDU is ON 
  if (  (writeDU == true ) && 
       (( bRunStrict == true ) || ( (bRunStrict == false) && (badDU == false)) )
     ) { 
  // END SPRx14

    fwrite((void *)rawdata, RawTM::XMCS::header_size, 1, outFile);
    fwrite((void *)(rawdata + RawTM::XMCS::header_size), DU->Size - RawTM::XMCS::header_size , 1, outFile);

  // SPR x14
  }
  // END SPRx14
  //////////////////////////////////////

  } // End of while fread TMDU
  
  // CRx16 we dump statistical information
  errstr << "------------------ DU READER ENDS ------------------" 
         << Sas::message(Msg::App, Msg::Verbose);

  errstr << "DU statistics information "
         << " Earliest Time: " << firstTime.toString(7)
         << " Last Time: " << lastTime.toString(7)
  //       << " Max DU size found: " << maxSize
         << Sas::message(Msg::App, Msg::Sparse);
  
  errstr << " TOTAL DUs found: " << nFrames
         << Sas::message(Msg::App, Msg::Sparse);
  
  errstr << "DU statistics information "
         << " Type 1(TM): " << aDUCounter[1]
         << " Type 2(reserved): " << aDUCounter[2]
         << " Type 3(TCH): " << aDUCounter[3]
         << " Type 4(MOUT): " << aDUCounter[4]
         << " Type 5(MOUTP): " << aDUCounter[5]
         << " Type 6(Timeline): " << aDUCounter[6]
         << " Type 7(TC): " << aDUCounter[7]
         << Sas::message(Msg::App, Msg::Sparse);
  // END CRx16 

 
}

void DUExtractor::processDU(FILE * rawFile, int *maxDU, bool *loop, bool bRunStrict ) {
  
  RawTM::XMCS::DataUnit *DU;

  unsigned char rawdata[RawTM::XMCS::max_size];
  
  STime saneMinTime = STime("1998-01-01T00:00:00.0");
  STime saneMaxTime = STime("2025-01-01T00:00:00.0");
  STime lastValidTime;

  // NEW SPRx11
  STime lastTime;
  STime firstTime;
  STime firstDUTime;
  int nFrames = 0;
  // END SPRx11

  bool firstDU = true;
  *loop = true;
  bool processDU = true;
  int maxSize = 0;
  lastDUTime = STime::earliest;
  lastSequenceCount = 0;

  // List of timestamps received
  std::set<signed long long> timeList;

  hasTimeFilter = true;
  
  BTI bti;

    // Reset error count
    int nErrors = 0;

    // Read the data
    while (fread((void *)rawdata, RawTM::XMCS::header_size, 1, rawFile)) {

      // NEW CR x11 Additional  tracing
      errstr << "------------------ NEW TMDU BEGINS ------------------" << Sas::message(Msg::App, Msg::Verbose); 

      processDU = true;
	  
      char isS2K = rawdata[15];
	  
      // Byte swap fields
      Utils::vmsByteSwap(&rawdata[0], 2); // Size

	  if (isS2K == 0x1) { // SCOS-I == 0xAA, SCOS-2000 == 0x1 
          errstr << "Is SCOS-2000 file" << Sas::message(Msg::App, Msg::Verbose);
    	  Utils::sunByteSwap(&rawdata[2], 4); // Time
    	  Utils::sunByteSwap(&rawdata[6], 4); // Time
    	  Utils::sunByteSwap(&rawdata[10], 4); // Counter
	  }
	  else {	  
          errstr << "Is SCOS-I file" << Sas::message(Msg::App, Msg::Verbose);
          Utils::vmsByteSwap(&rawdata[2], 8); // Time
          Utils::vmsByteSwap(&rawdata[10], 4); // Counter
      }
	  
      // Cast header onto header record
      DU = (RawTM::XMCS::DataUnit *)rawdata;

	STime duTime;
         // Get the DU time
	  if (isS2K == 0x1) {	  
         duTime = Utils::readTDSTime(DU->TDSTime);
      }
	  else {
         duTime = Utils::readVMSTime(DU->VMSTime);
	  }


      // NEW CR x11 Additional  tracing
      errstr << "DU Time information:" << Sas::message(Msg::App, Msg::Verbose); 
      errstr <<  hex << DU->VMSTime 
	  	<< " " 
	  	<< DU->TDSTime.Seconds 
	  	<< " " 
		<< DU->TDSTime.Microseconds 
	  	<< " " << dec
		<< " Type: " 
		<< (int)DU->Type 
	  	<< " " 
		<< duTime.toString(6) 
		<< Sas::message(Msg::App, Msg::Verbose);
      // END CR x11

      // NEW SPRx11
      if (nFrames == 0) firstTime=duTime;

      if (firstDU) {
         firstDUTime=duTime;
      }
      else {
         lastTime=duTime;
      }
      nFrames++;
      // END SPRx11

      // Check for sane length
      if (DU->Size < 17) {

        // NEW SPR x7 
        //if ( !bStrictRun ) {
        if ( !bRunStrict ) {
	errstr << "WARNING Invalid DU found, MAY ABORT but ignore and jump next file " 
	       << " Type: " << (int)DU->Type 
	       << " Size: " << dec << DU->Size 
	       << " Counter: " << hex << DU->Counter << dec
	       << " Last valid DU: " << lastValidTime.toString(7)
	       << Sas::warning("duInvalidSize");
	  break;
        
        } else {
	errstr << "Invalid DU found, ABORT" 
	       << " Type: " << (int)DU->Type 
	       << " Size: " << dec << DU->Size 
	       << " Counter: " << hex << DU->Counter << dec
	       << " Last valid DU: " << lastValidTime.toString(7)
	       << Sas::error("duInvalidSize");
        }
	
      }
      
      // Read payload
      fread((void *)(rawdata + RawTM::XMCS::header_size), DU->Size - RawTM::XMCS::header_size , 1, rawFile);
    
      if ((DU->Type < 1) || (DU->Type > 7) ||
	  (duTime < saneMinTime) || (duTime > saneMaxTime)) {

	errstr << "Invalid DU found, trying to resync." 
	       << " Counter: " << hex << DU->Counter << dec
	       << " Expected Counter: " << hex << lastSequenceCount + 1 << dec
	       << Sas::warning("duInvalid");

	nErrors++;
	continue;

      }

      if (nErrors > 2) {
	errstr << "Too many bad frames, Last valid DU: " 
	       << lastValidTime.toString(7) << Sas::error("duAbort");
      }

      // If the DU is earlier than we want, skip to next
      if (duTime < startTime) {

        // NEW CR x11
        errstr << "DU time "
               << duTime 
               << " earlier that startTime " 
               << startTime 
               << ": continuing " 
               << Sas::message(Msg::App, Msg::Verbose);
        // END CR x11

	continue;
      }

      // If the DU is later than we want, skip to next file
      if (duTime > stopTime) {
	  *loop = false;
          // NEW CR x11
          errstr << "DU time "
                 << " later than stopTime " 
                 << stopTime
                 << ": aborting this file and stopping looping " 
                 << Sas::message(Msg::App, Msg::Sparse);
          // END CR x11
	  break;
      }

      // THIS IS HARCODED AS true ALWAYS!!!!
      if (hasTimeFilter) {

	bool accept = true;

        // NEW CRx20
        std::vector<int16>::const_iterator gsIDiter = bti.bti_gsid_Filter.begin();
        // END CRx20

        for (std::vector<BTI_pair>::const_iterator timeIter = bti.timeFilter.begin();
	     timeIter!= bti.timeFilter.end();
	     //timeIter++) {
	     timeIter++, gsIDiter++ ) {
		 
          // NEW CRx20 Now should avoid using GSID entries
	  //if ((duTime > (*timeIter).first) && (duTime < (*timeIter).second)) {
	  if ( (duTime > (*timeIter).first) && 
               (duTime < (*timeIter).second) &&
               ((*gsIDiter) == 0 ) 
               ) {
            errstr << "WARNING DU is ignored according BTI file time info." 
	      << " DU Time: " << duTime.toString(7)
	      << Sas::message(Msg::App, Msg::Sparse);
          // END CRx20
	    accept = false;
	    break;
	  }

	}
	
	if (!accept) {
	  continue;
	}

      }

      if (DU->Size > maxSize) {
	maxSize = DU->Size;
      }

      errstr << "DU of" 
	     << " Type: " << (int)DU->Type 
	     << " Size: " << dec << DU->Size 
	     << " Counter: " << hex << DU->Counter << dec
	     << " Time: " << duTime.toString(7)
	     << Sas::message(Msg::App, Msg::Verbose);
      
      // Only perform checks on subsequent iterations
      if (!firstDU) {
	
	// Check for data gaps and repeat frames from merged CDs
	if ((DU->Counter != lastSequenceCount + 1) && (lastSequenceCount != 0xFFFFFFFF)) {
	  
	  if (DU->Counter == 0) {
	    
	    errstr << "Counter reset, expected " 
		   << hex << lastSequenceCount + 1 << dec
		   << Sas::warning("counterReset");
	    
	  }
	  else {

	    // Check we never get out-of-sequence frames with the same timestamp
	    // This will catch repeat frames with a lower counter
	    if ( DU->Counter < lastSequenceCount ){

	      if (timeList.find(DU->VMSTime) != timeList.end()) {
		errstr << "Repeat frame detected: found " 
		       << hex << DU->Counter
		       << " expected " 
		       << lastSequenceCount + 1 << dec
		       << Sas::warning("duRepeat");	
		
		processDU = false;
		
	      }

	    }
	    else {

	      // In the case you get a much higher counter (delta > 10) and a repeated timestamp
	      // it's a duplicate
	      if (( DU->Counter - lastSequenceCount > 10) && (timeList.find(DU->VMSTime) != timeList.end())) {

		errstr << "Repeat frame detected: found " 
		       << hex << DU->Counter
		       << " expected " 
		       << lastSequenceCount + 1 << dec
		       << Sas::warning("duRepeat");	
		
		processDU = false;
	      
	      }
	      else {

		errstr << "Data gap detected: found " 
		       << hex << DU->Counter
		       << " expected " 
		       << lastSequenceCount + 1 << dec
		       << Sas::warning("duDataGap");
                // SPRx32
                errstr << "DUExtractor: gap detected at " 
                     << duTime.toString(7)
                     << Sas::message(Msg::App, Msg::Verbose);

                std::cout <<  "DUExtractor: gap detected at "
                     << duTime.toString(7) << std::endl;
		
	      }
	    }
	  }
	}
      }

      if (processDU) {

	lastValidTime = duTime;

	// Insert timestamp into list
	timeList.insert(DU->VMSTime);

	// Remember values
	lastSequenceCount = DU->Counter;
	lastDUTime = duTime;
	
	// Send DU to the processors
	for(unsigned int i = 0; i < Listeners[(DUProcessor::type)DU->Type].getMap().size(); i++) {	
	  
	  DUProcessor *p = (DUProcessor *)Listeners[(DUProcessor::type)DU->Type].getMap()[i];
	  p->extractData(DU);	
	  
	}	  
	
      }

      // Flip first DataUnit flag
      if (firstDU == true) {
	firstDU = false;
      }

    }


    // We have ended the processing, now returns
    // the maximun size of DU found
    *maxDU = maxSize;

   // NEW SPRx11
   // Finally, we dump the check information
   errstr << "------------------ TMDU PROCESSOR ENDS ------------------" << Sas::message(Msg::App, Msg::Verbose); 

   errstr << "DU checking information " 
	     << " Earliest Time: " << firstTime.toString(7)
	     << " First Processed Time: " << firstDUTime.toString(7)
	     << " Last Time: " << lastTime.toString(7)
	     << " Last Processed Time: " << lastDUTime.toString(7)
	     << " Max DU size found: " << maxSize 
	     << " N of Frames found: " << nFrames
	     << Sas::message(Msg::App, Msg::Sparse);
   // END SPRx11   
    
}

