// CRx17    jcv 29-Jun-09 Original new duproc design 

#include <iostream>
#include <utility>
#include <iterator>
#include <set>
#include <fstream>

#include "errstr.h"
#include "param.h"

#include "ProcessorList.h"
#include "DataDefinition.h"
#include "DUStream.h"
#include "DUProcessor.h"
#include "Utils.h"
#include "SasFileLocator.h"	

#include "FilterDefinition.h"

DUstream::DUstream() {

  string sFullName;
  string sCacheDir = stringParameter("cachedir");
  string sRawFile = stringParameter("rawfile");

  // open the rawfile (always)
  sFullName = sCacheDir;
  sFullName.append("/");
  sFullName.append(sRawFile);
  errstr << "Full filename to read " << sFullName << Sas::message(Msg::App, Msg::Verbose);

  if ((pRawFile = fopen(sFullName.c_str(),"r")) == NULL) {
      errstr << "Failed to open data file " << sFullName
             << Sas::fatal("failOpenDatafile");
  }


}

DUstream::~DUstream() {

 fclose(pRawFile);

}

void DUstream::registerProcessor(DUProcessor *proc) {

  Listeners[proc->processorType()].addProcessor(proc);

}

void DUstream::unregisterProcessor(DUProcessor *proc) {

  Listeners[proc->processorType()].removeProcessor(proc);

}



bool DUstream::FilterData(RawTM::XMCS::DataUnit * DU, STime duTime, bool badDU) {

  bool bDummie;

  bDummie = badDU;

  // Verbose output  
  errstr << "NO Filter for DU of" 
             << " Type: " << (int)DU->Type 
             << " Size: " << dec << DU->Size 
             << " Header size " << RawTM::XMCS::header_size 
             << " Counter: " << hex << DU->Counter << dec
             << " Time: " << duTime.toString(7)
             << Sas::message(Msg::App, Msg::Verbose);

  

  return true;

}

void DUstream::ProcessData(RawTM::XMCS::DataUnit * DU,  unsigned char *rawdata) {

  // Verbose output  
  errstr << "VOID Process for DU of" 
             << " Type: " << (int)DU->Type 
             << " Size: " << dec << DU->Size 
             << " Header size " << RawTM::XMCS::header_size 
             << " Counter: " << hex << DU->Counter << dec
             << Sas::message(Msg::App, Msg::Verbose);

  errstr << "Header is " << hex 
         << rawdata[0] << dec << Sas::message(Msg::App, Msg::Verbose);

}

void DUstream::streamDU() {

  RawTM::XMCS::DataUnit *DU;

  unsigned char rawdata[RawTM::XMCS::max_size];
  
  STime saneMinTime = STime("1998-01-01T00:00:00.0");
  STime saneMaxTime = STime("2025-01-01T00:00:00.0");
  STime lastValidTime;

  STime lastTime;
  STime firstTime;

  int aDUCounter[8];
  int nFrames = 0;

  bool badDU = false;
  bool bIsProcessable = true;

  lastDUTime = STime::earliest;
  lastSequenceCount = 0;

  // List of timestamps received
  std::set<signed long long> timeList;

  // Statistics on DU types
  for (unsigned int i=0; i<8; i++) aDUCounter[i]=0; 

  ////////////////////////////////////////////////////////
  // Read the data: TMDU Frame Header
  while (fread((void *)rawdata, RawTM::XMCS::header_size, 1, pRawFile)) {

  nFrames++;
  errstr << "--------------DU " << nFrames 
         << " READ -----------" << Sas::message(Msg::App, Msg::Verbose); 

  badDU = false;

  char isS2K = rawdata[15];

  // Byte swap fields
  Utils::vmsByteSwap(&rawdata[0], 2); // Size

  if (isS2K == 0x1) { // SCOS-I == 0xAA, SCOS-2000 == 0x1 
          errstr << "DU type: SCOS-2000" << Sas::message(Msg::App, Msg::Verbose);
          Utils::sunByteSwap(&rawdata[2], 4); // Time
          Utils::sunByteSwap(&rawdata[6], 4); // Time
          Utils::sunByteSwap(&rawdata[10], 4); // Counter
          }
          else {
          errstr << "DU type: SCOS-I" << Sas::message(Msg::App, Msg::Verbose);
          Utils::vmsByteSwap(&rawdata[2], 8); // Time
          Utils::vmsByteSwap(&rawdata[10], 4); // Counter
  }

  // Cast header onto header record
  DU = (RawTM::XMCS::DataUnit *)rawdata;

  // Checks 
  if (DU->Size < 17) {
   errstr << "WARNING Invalid DU Size found" << DU->Size  << Sas::warning("duInvalidSize");
   errstr << "WARNING Invalid DU Size found" << DU->Size  << Sas::message(Msg::App, Msg::Verbose);

  badDU = true;

  }

  if ((DU->Type < 1) || (DU->Type > 7)) {
   errstr << "WARNING Invalid DU Type found" << DU->Type << Sas::warning("duInvalidType");
   errstr << "WARNING Invalid DU Type found" << DU->Type << Sas::message(Msg::App, Msg::Verbose);

  badDU = true;

  }
  // Statistics on DU Type
  else
     aDUCounter[DU->Type]++;

  // Get the DU time
  STime duTime;
  if (isS2K == 0x1) {     
         duTime = Utils::readTDSTime(DU->TDSTime);
      }
  else {
         duTime = Utils::readVMSTime(DU->VMSTime);
  }

  if ((duTime < saneMinTime) || (duTime > saneMaxTime)) {
   errstr << "WARNING Invalid DU Sane Time found" << Sas::warning("duInvalidSane");
   errstr << "WARNING Invalid DU Sane Time found" << Sas::message(Msg::App, Msg::Verbose);

  badDU = true;

  }

  // Additional statistics
  if (nFrames == 1) firstTime=duTime;
  lastTime=duTime;


  // Verbose output  
  errstr << "DU of" 
             << " Type: " << (int)DU->Type 
             << " Size: " << dec << DU->Size 
             << " Header size " << RawTM::XMCS::header_size 
             << " Counter: " << hex << DU->Counter << dec
             << " Time: " << duTime.toString(7)
             << Sas::message(Msg::App, Msg::Verbose);

  ////////////////////////////////////////////////////////
  // Payload processing
  // Read the data: read the payload
  fread((void *)(rawdata + RawTM::XMCS::header_size), DU->Size - RawTM::XMCS::header_size , 1, pRawFile);

  // Insert timestamp into list
  timeList.insert(DU->VMSTime);

  // Statistics here
  lastDUTime = duTime;

  /////////////////////////////////////////////////////////////////
  // Dump the TMDU
  // RAWDATA HAS BEEN MODIFIED
  /////////////////////////////////////////////////////////////////

  // Byte swap fields
  Utils::vmsByteSwap(&rawdata[0], 2); // Size

  if (isS2K == 0x1) { // SCOS-I == 0xAA, SCOS-2000 == 0x1 
      errstr << "Undoing SCOS-2000 " << Sas::message(Msg::App, Msg::Verbose);
      Utils::sunByteSwap(&rawdata[2], 4); // Time
      Utils::sunByteSwap(&rawdata[6], 4); // Time
      Utils::sunByteSwap(&rawdata[10], 4); // Counter
          }
          else {
      errstr << "Undoing SCOS-I " << Sas::message(Msg::App, Msg::Verbose);
      Utils::vmsByteSwap(&rawdata[2], 8); // Time
      Utils::vmsByteSwap(&rawdata[10], 4); // Counter
  }

  // End of swaps

  //////////////////////////////////////

  bIsProcessable=FilterData(DU, duTime, badDU);

  if (bIsProcessable)  ProcessData(DU,rawdata);

  //////////////////////////////////////

  } // End of while fread TMDU
  
  // we dump statistical information
  errstr << "------------------ DU READER ENDS ------------------" 
         << Sas::message(Msg::App, Msg::Verbose);

  errstr << "DU statistics information "
         << " Earliest Time: " << firstTime.toString(7)
         << " Last Time: " << lastTime.toString(7)
  //       << " Max DU size found: " << maxSize
         << Sas::message(Msg::App, Msg::Sparse);
  
  errstr << " TOTAL DUs found: " << nFrames
         << Sas::message(Msg::App, Msg::Sparse);
  
  errstr << "DU statistics information "
         << " Type 1(TM): " << aDUCounter[1]
         << " Type 2(reserved): " << aDUCounter[2]
         << " Type 3(TCH): " << aDUCounter[3]
         << " Type 4(MOUT): " << aDUCounter[4]
         << " Type 5(MOUTP): " << aDUCounter[5]
         << " Type 6(Timeline): " << aDUCounter[6]
         << " Type 7(TC): " << aDUCounter[7]
         << Sas::message(Msg::App, Msg::Sparse);
 
}

