// CRx17    jcv 29-Jun-09 Original new duproc design 
// SPRx18   jcv 28-Jul-09 Refined filtering when bad DU are found 

#include <iostream>
#include <utility>
#include <iterator>
#include <set>
#include <fstream>

#include "errstr.h"
#include "param.h"

//#include "ProcessorList.h"
#include "DataDefinition.h"
#include "DUwriter.h"
//#include "DUProcessor.h"
#include "Utils.h"
#include "SasFileLocator.h"	

#include "FilterDefinition.h"

DUwriter::DUwriter() {

  string sFullName;

  AppliedFilterInfo.bFilterRequired=false;
  AppliedFilterInfo.bRemoveRequired=false;

  pOutFile = NULL;

    string sOutFile = stringParameter("outfile");
    bool bFilterDU = booleanParameter("filterdu");

    errstr << "Data will be written to :" << sOutFile << Sas::message(Msg::App, Msg::Verbose);

    bWriteBadData = booleanParameter("baddata");
    startTime = stringParameter("starttime");
    stopTime  = stringParameter("endtime");

    errstr << "Bad Data will be dump? (1=yes):" << bWriteBadData 
           << Sas::message(Msg::App, Msg::Verbose);
    errstr << "DUs will be written if stamped within " 
           << startTime.toString(7)
           << " and " << stopTime.toString(7)
           << Sas::message(Msg::App, Msg::Verbose);

    sFullName = sCacheDir;
    sFullName.append("/");
    sFullName.append(sOutFile);
    errstr << "Full filename to write " << sFullName << Sas::message(Msg::App, Msg::Verbose);

    if ((pOutFile = fopen(sFullName.c_str(),"wb")) == NULL) {
      errstr << "Failed to open outputfile " << sFullName
             << Sas::fatal("failOpenOutfile");
    }

    // read the filterfile, if needed
    if (bFilterDU)  {
      string sFilterFile = stringParameter("filterfile");
      errstr << "DUs will be filtered." << Sas::message(Msg::App, Msg::Sparse);
      errstr << "DU filter file: " << sFilterFile << Sas::message(Msg::App, Msg::Verbose);
      this->readFilterFile(sFilterFile,&AppliedFilterInfo);
    } // End if bFilterDU file needed


}

DUwriter::~DUwriter() {

 if (pOutFile != NULL) 
    fclose(pOutFile);


}


void DUwriter::readFilterFile(string sFilename, FilterInfo * AppliedFilters) {

  // Filter Fields /////////////
  string sDummie;
  string sDirective;

  // Fields for DIRECTIVE-B, DU removal  based in Time and Type 
  string sRemovalStart;
  string sRemovalEnd;
  string sKeepDU;

  string sLocatedFile = fileLocator.locate(sFilename, "", false, true);
  if (sLocatedFile.length() == 0)
    {
      errstr << "file not found (check fullpath?). " << Sas::error("NotFilterFile");
      exit(1);
    }

  ifstream ifs(sLocatedFile.c_str());
  if (! ifs.good()) {
         errstr << "file stream invalid. " << Sas::error("BadStream");
         exit(1);
  }

  try {
        getline(ifs, sDummie); 
        getline(ifs, sDirective); 

        if (sDirective.find("DIRECTIVE-B-") != string::npos) {
         // DIRECTIVE B, Start-End Time DU removal 
         errstr << "DIRECTIVE B, Start-End Time DU removal" 
                     << Sas::message(Msg::App, Msg::Verbose); 
         getline(ifs, sRemovalStart); 
         sRemovalStart=sRemovalStart.substr(10);
         getline(ifs, sRemovalEnd); 
         sRemovalEnd=sRemovalEnd.substr(10);
         getline(ifs, sKeepDU); 
         sKeepDU=sKeepDU.substr(10);

         AppliedFilters->bRemoveRequired=true;
         AppliedFilters->RemoveEnd=sRemovalEnd;
         AppliedFilters->RemoveStart=sRemovalStart;
         AppliedFilters->KeepDUTypes=sKeepDU;

        } // End directive branching 
        else {
         errstr << "DIRECTIVE B Not found" << Sas::fatal("directiveBnotfound"); 
        }
  } catch(...) {  // The try failed
        errstr << "Syntax err in filter file." << Sas::error("SyntaxFile");
        exit(1);
  }

}



bool DUwriter::FilterData(RawTM::XMCS::DataUnit * DU, STime duTime, bool badDU) {

  bool bIsProcessable = true;
  bool bDUKeep = true;

  // Verbose output  
  errstr << "Ready to filter DU of" 
             << " Type: " << (int)DU->Type 
             << " Size: " << dec << DU->Size 
             << " Header size " << RawTM::XMCS::header_size 
             << " Counter: " << hex << DU->Counter << dec
             << " Time: " << duTime.toString(7)
             << Sas::message(Msg::App, Msg::Noisy);

  // If the removal is activated, check and flag properly
  if (AppliedFilterInfo.bRemoveRequired)  
    {
    // Added removal by checking the DU type
    // If not found in the filter string, then do not keep it 

    std::ostringstream ostrDUType;
    ostrDUType << (int)DU->Type;

    errstr << " Filter file activated. Searching  DU type " << ostrDUType.str() 
           << " in filter to-be-kept string " << AppliedFilterInfo.KeepDUTypes
           << Sas::message(Msg::App, Msg::Noisy);

    if ( AppliedFilterInfo.KeepDUTypes.find(ostrDUType.str()) == string::npos) 
       {
         errstr << " Not found in string" << Sas::message(Msg::App, Msg::Noisy);

         // Not found, so candidate for being removed 
         if (  (duTime > AppliedFilterInfo.RemoveStart ) &&
          (duTime < AppliedFilterInfo.RemoveEnd ) ) {
            // Remove it when within given filter start-end span
            bDUKeep = false;
            errstr << "This DU will be REMOVED!!!" << Sas::message(Msg::App, Msg::Verbose);
         }

       } else {
         errstr << " found in string" << Sas::message(Msg::App, Msg::Noisy);
       }

  } // End If filters required

  // It will only dump the DU if , 
  // It has not been filtered out i.e. bDUKeep is ON and 
  if (  (bDUKeep == true ) && 
  // duTime is within the start-end time
   (  (duTime > startTime ) && (duTime < stopTime )  ) &&
  // we dump both good and bad data, i.e. bWriteBadData = yes, or
  // we dump just good data, i.e badDU = no (i.e. and bWriteBadData = no)
       (( bWriteBadData == true ) || ( (bWriteBadData == false) && (badDU == false)) )
     ) { 
          bIsProcessable = true;
       } else
       {
          bIsProcessable = false;
       }

   errstr << " DU filter returns: is processable (1=yes) " << bIsProcessable 
          << Sas::message(Msg::App, Msg::Noisy);
   return bIsProcessable;

}

void DUwriter::ProcessData(RawTM::XMCS::DataUnit * DU,  unsigned char *rawdata) {

    fwrite((void *)rawdata, RawTM::XMCS::header_size, 1, pOutFile);
    fwrite((void *)(rawdata + RawTM::XMCS::header_size), DU->Size - RawTM::XMCS::header_size , 1, pOutFile);

}
