// SPRx22 jcv 4-Feb-2010 gcc4 migration
// CRx19 jcv 31-Aug-2010 New ground station Kourou Galliot

#ifndef DATAUNIT_H
#define DATAUNIT_H

#include "STime.h"
#include "Utils.h"

namespace RawTM {

  namespace XMM {
    
    static const unsigned int max_size = 518;
    
    typedef struct {
      union {
	unsigned char	rawPacketID[4];
	struct {
	  unsigned int	SequenceCount		:14;
	  unsigned int	SeqmentationFlag	:2;
	  unsigned int	APID			:11;
	  unsigned int	DataFieldHeaderFlag	:1;
	  unsigned int	Type			:1;
	  unsigned int	VersionNumber		:3;
	} PacketID;
      };
      union {
	unsigned char   rawPacketLength[2];
	unsigned short	PacketLength;
      };
    } __attribute__ ((packed)) PacketHeader;
    
    typedef struct {
      union {
	union {
	  struct
	  {
	    unsigned short	PacketSubtype	:4;
	    unsigned short	PacketType	:4;
	    unsigned short	ChecksumFlag	:2;
	    unsigned short	Spare		:6;
	  } __attribute__ ((packed)) TM;
	  unsigned char           rawTM[2];
	};
	union {
	  struct
	  {
	    unsigned short	PacketSubtype	:4;
	    unsigned short	PacketType	:4;
	    unsigned short	ChecksumFlag	:2;
	    unsigned short	Spare		:6;
	    unsigned int        CoarseTime;
	    unsigned short      FineTime;
	  } __attribute__ ((packed)) TMTime;
	  struct {
	    unsigned char           rawData[2];
	    unsigned char           rawCoarse[4];
	    unsigned char           rawFine[2];
	  } __attribute__ ((packed)) rawTMTime;
	};

	union {
	  struct {
	    unsigned int        i1;
	    unsigned int        i2;
	  } RawTime;
	  
	  struct {
	    unsigned char        c1;
	    unsigned char        c2;
	    unsigned short       s1;
	    unsigned short       s2;
	    unsigned short       s3;
	  } RawTime2;
	  
	  struct {
	    unsigned char	      SField;
	    unsigned char	      PField;
	    unsigned int	      Coarse;
	    unsigned short            Fine;
	  } Time;
	};
      };
    } __attribute__ ((packed)) DataFieldHeader;

    typedef struct {
      union {
	struct {
	  PacketHeader	        Header;
	  DataFieldHeader	DFHeader;
	};
	unsigned char	Data[ max_size ];
      };
    } __attribute__ ((packed)) Packet;

  };

  namespace NCTRS {

    static const unsigned int data_size = 882;
    static const unsigned int max_packets = 200;	

    enum frameType {VC0 = 0, VC7 = 7};
    enum quality {good = 0, bad = 1};
    enum dsType { SLC = 0, MC = 1, VC = 2, BAD = 3 };

    // NEW CRx19 New ground station Kourou-Galliot
    //enum gsID { VIL1 = 5, REDU = 10, VIL2 = 13, KRU = 21, PER = 22, NNO = 23, AGO = 41, DON = 44 };
    enum gsID { VIL1 = 5, REDU = 10, VIL2 = 13, KRU = 21, PER = 22, NNO = 23, AGO = 41, DON = 44, KGAL = 49 };
    
    typedef struct {
      union {
	unsigned char   rawFrameID[2];
	struct {
	  unsigned short	OpControl	:1;
	  unsigned short	VCID		:3;
	  unsigned short	SCID		:10;
	  unsigned short	Version		:2;
	} FrameID;
      };
      unsigned char	MFC;
      unsigned char	VFC;
      union {
	unsigned char   rawFrameStatus[2];
	struct {
	  unsigned short	FirstHeader	:11;
	  unsigned short	SegID		:2;
	  unsigned short	PacketOrder	:1;
	  unsigned short	SyncFlag	:1;
	  unsigned short	SecHeaderFlag	:1;
	} FrameStatus;
      };
      unsigned char	Data[ data_size ];
      union {
	unsigned char   rawCLCW[4];
	unsigned long	CLCW;
      };
    } __attribute__ ((packed)) TransferFrame;
    
    typedef struct 
    {
      union {
	unsigned char   rawPacketSize[4];
	unsigned int	PacketSize;
      };
      union {
	unsigned char   rawSCID[2];
	unsigned short	SCID;
      };
      unsigned char	DSType;
      unsigned char	VCID;
      union {
	unsigned char   rawGSID[2];
	unsigned short	GSID;
      };
      union {
	struct {
	  unsigned char   Days[2];
	  unsigned char   Millis[4];
	  unsigned char   Micros[2];
	} rawTime;
	Utils::CDSTime	Time;
      };
      union {
	unsigned char   rawQuality[2];
	unsigned short	Quality;
      };
    } __attribute__ ((packed)) FrameHeader;

    typedef struct {
      FrameHeader       Header;
      TransferFrame	Frame;
      STime             CorrectedFrameTime;

    // SPRx22 gcc4 migration ignored the packet attrb as Stime is by itlsef notpacket
    //} __attribute__ ((packed)) TMDU;
    } TMDU;
          
  };

  namespace XMCS {
    
    // Header Size
    static const unsigned int header_size = 16;
    static const unsigned int max_size = 65535;

    // Types of data we expect (FIXME when XSCS-2000 is delivered)
    enum type {TM = 1, TC = 3, MOUT = 4, MOUTP = 5, TLEX = 6, TCORR = 7, nTypes = 8};
    
    typedef struct {
      unsigned short	Size;
	  union {
	      signed long long	VMSTime;
		  Utils::tdsTime    TDSTime;
	  };  
      unsigned long	Counter;
      unsigned char	Type;
      unsigned char	Reserved;
      unsigned char     Data [ max_size - header_size ];
    } __attribute__ ((packed)) DataUnit;

  };

 
};

#endif
