// CR x11   jcv           Fixed output tracing information
// CRx14    jcv 20-Apr-09 Infrastructure for dump/fixer
// SPRx14   jcv 22-Apr-09 Enhancements for packet display
// CRx15    jcv 05-May-09 Filter capabilities
// SPRx16   jcv 20-May-09 DFH Capabilities
// SPRx32   jcv 07-May-10 Replaced duDataGaps by Idle warning 
// CRx19   jcv 31-Aug-10 New ground station Kourou-Galliot
// CRx20   jcv 21-Sep-10 Added bti filtering per g/s id 
// CRx21   jcv 27-Sep-10 Additional hex dump of Frame header

#include "NCTRSProcessor.h"
#include "PacketProcessor.h"
#include "DataDefinition.h"
#include "Utils.h"

#include "errstr.h"

// NEW CRx10
#include "param.h"

// CRx15
#include "FilterDefinition.h"

// SPRx16
#include "dfhtimestamp.h"

// NEW CRx14
void NCprintHex(unsigned char *buf, int len);
// END CRx14

// CRx20
#include "CalOalUtility.h"
// END CRx20


NCTRSProcessor::NCTRSProcessor(FrameConsumer *proc, NCTRSProcessor::VC c) : dataProc(proc), channel(c) {
  
  isFirstFrame = true;
  isOverspill = false;
  iCopied = 0;

  refXMCSCorrection = STime("2002-01-22T12:00:00.0");
 
  // CRx20
  // Load delays into memory (and boll about if use it (or not)
  bFullBTI=booleanParameter("fullBTI");

  DataSet *dataset = dataSetServer->open("bti.fit", Dal::Read);

  Table *tab = dataset->table("BTI");

  Column *startCol = tab->column("START");
  Column *endCol   = tab->column("END");
  Column *GSCol   = tab->column("GSID");

  ColumnData *startColD = startCol->data();
  ColumnData *endColD   = endCol->data();
  ColumnData *GSColD = GSCol->data();
  int16 *iGSid = GSColD->int16Data();

  for (unsigned int i = 0; i < tab->rows(); i++) {

    STime startTime(CalOalUtility::stringColumnRow(startColD, i));
    STime endTime(CalOalUtility::stringColumnRow(endColD, i));
    BTI_pair entry(startTime, endTime);

    timeFilter.push_back(entry);
    bti_gsid_Filter.push_back(iGSid[i]); 

  }

  dataSetServer->close(dataset);

 // END CRx20 
 
}

// CRx14
// CRx15
//void NCTRSProcessor::procData(RawTM::XMCS::DataUnit *DU){
// SPRx16
//void NCTRSProcessor::procData(RawTM::XMCS::DataUnit *DU, FilterInfo *Filters){
void NCTRSProcessor::procData(RawTM::XMCS::DataUnit *DU, 
                              FilterInfo *Filters,
                              int iNumGeneric, int iNumSpecific,
                              VTMZZ_TNODFHTIME *MyarNoDFHTime){

  // CRx20
  errstr << "WARNING: Syntax calling obsolete function" 
   << Sas::warning("Obsolete");
  
  // Make a copy of the Raw DU data
  if ( DU->Size - RawTM::XMCS::header_size + sizeof(STime) != sizeof(RawTM::NCTRS::TMDU)) {
    errstr << "NCTRS Fixer: XMM DU with NCTRS DU of wrong size!"
	   << Sas::warning("WrongSize");
    return;
  }    

  ////////////////////////////////////////
  // Performs a memcpy i.e. Original data will NOT be modified
  memcpy(&nctrsDU, DU->Data, DU->Size - RawTM::XMCS::header_size);
  
  // Byte swap NCTRS DU 
  Utils::sunByteSwap(nctrsDU.Header.rawPacketSize, sizeof(nctrsDU.Header.rawPacketSize));
  Utils::sunByteSwap(nctrsDU.Header.rawSCID, sizeof(nctrsDU.Header.rawSCID));
  Utils::sunByteSwap(nctrsDU.Header.rawGSID, sizeof(nctrsDU.Header.rawGSID));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Days, sizeof(nctrsDU.Header.rawTime.Days));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Millis, sizeof(nctrsDU.Header.rawTime.Millis));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Micros, sizeof(nctrsDU.Header.rawTime.Micros));
  Utils::sunByteSwap(nctrsDU.Header.rawQuality, sizeof(nctrsDU.Header.rawQuality));

  errstr << "NCTRS Fixer: NCTRS Header: " 
	 << " FrameSize " << nctrsDU.Header.PacketSize 
	 << " SCID " << nctrsDU.Header.SCID 
	 << " DSType " << (int)nctrsDU.Header.DSType 
	 << " VCID " << (int)nctrsDU.Header.VCID 
	 << " GSID " << nctrsDU.Header.GSID 
	 << " Quality " << nctrsDU.Header.Quality 
	 << " Time " 
	 << nctrsDU.Header.Time.Days << ":"
	 << nctrsDU.Header.Time.Milliseconds << ":"
	 << nctrsDU.Header.Time.Microseconds             
	 << Sas::message(Msg::App, Msg::Verbose);

  if (nctrsDU.Header.Quality == RawTM::NCTRS::bad) {

    errstr << "NCTRS Fixer: Bad quality NCTRS frame"
	   << Sas::warning("BadFrame");
    
    return;
    
  }

  if ( nctrsDU.Header.SCID != 193 ) {
    
    errstr << "NCTRS Fixer: Data not from S/C: S/C ID = " << nctrsDU.Header.SCID
	   << Sas::warning("BadSC");
    
    return;

  }

  if ( (nctrsDU.Header.GSID != RawTM::NCTRS::VIL1 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::REDU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::VIL2 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::KRU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::PER ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::NNO ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::AGO ) &&
       // NEW CRx19
       (nctrsDU.Header.GSID != RawTM::NCTRS::KGAL ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::DON ) ) {
    errstr << "NCTRS Fixer: WARNING Bad GS ID" << nctrsDU.Header.GSID
	   << Sas::warning("BadGS");
    
    return;

  }
 
  // CRx21
  // Print 6 bytes, in hexadecimal
  errstr << "NCTRS Fixer: NCTRS Transfer Frame (Prim. Header): "
    << hex
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameID[0]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameID[1]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.MFC
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.VFC
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameStatus[0]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameStatus[1]
    << dec
    << Sas::message(Msg::App, Msg::Verbose);
  // END CRx21
 
  // Byte swap Transfer Frame
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameID, sizeof(nctrsDU.Frame.rawFrameID));
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameStatus, sizeof(nctrsDU.Frame.rawFrameStatus));
  Utils::sunByteSwap(nctrsDU.Frame.rawCLCW, sizeof(nctrsDU.Frame.rawCLCW));

  errstr << "NCTRS Fixer: NCTRS Transfer Frame: " 
	 << " OpControl " << nctrsDU.Frame.FrameID.OpControl     
	 << " VCID " << nctrsDU.Frame.FrameID.VCID	  
	 << " SCID " << nctrsDU.Frame.FrameID.SCID          
	 << " Version " << nctrsDU.Frame.FrameID.Version	  
	 << " MFC " << (int)nctrsDU.Frame.MFC           
	 << " VFC " << (int)nctrsDU.Frame.VFC           
	 << " FirstHeader " << nctrsDU.Frame.FrameStatus.FirstHeader   
	 << " SegID " << nctrsDU.Frame.FrameStatus.SegID	 
	 << " PacketOrder " << nctrsDU.Frame.FrameStatus.PacketOrder   
  	 << " SyncFlag " << nctrsDU.Frame.FrameStatus.SyncFlag      
	 << " SecHeaderFlag " << nctrsDU.Frame.FrameStatus.SecHeaderFlag 
	 << " CLCW " << hex << nctrsDU.Frame.CLCW << dec 
   	 << Sas::message(Msg::App, Msg::Verbose);

 ////////////////////////////////////////////////////////
 // Once the frame is identified, the packets are extracted

  // SPRx14 We will need correct the Frame Time for having the packet time
  nctrsDU.CorrectedFrameTime = correctFrameTime(nctrsDU.Header.Time, nctrsDU.Header.GSID);

  // If this DU is on the right channel
//  if (nctrsDU.Frame.FrameID.VCID == channel) {

    RawTM::XMM::PacketHeader pHeader;
    unsigned int iLeft;
    
    unsigned int iOffset = nctrsDU.Frame.FrameStatus.FirstHeader;
    nPackets = 0;
    
    // BEGIN Packet Extraction goes here
    if (isFirstFrame) {
      errstr << "NCTRS Fixer: Processing first frame" 
             << Sas::message(Msg::App, Msg::Noisy);
      isFirstFrame = false;
    }
    else if (isOverspill) {
      errstr << "NCTRS Fixer: Overspill to be processed" 
             << Sas::message(Msg::App, Msg::Noisy);
      
      if (nctrsDU.Frame.FrameID.VCID != lastVC) {
        errstr << "NCTRS Fixer: VC mutated from last frame VC!"
               << Sas::warning("WrongVC");
      }
      else if ( (iOffset + iCopied) <= sizeof( RawTM::XMM::Packet ) ) {
        // Copy the overspill into the new packet
        extractedPackets[nPackets] = overspill;
        // iOffset gives the number of bytes not transmitted in the packet from the last frame

        // THE COPY PROCESS
        memcpy((void *)&extractedPackets[nPackets].Data[iCopied], 
               (void *)&nctrsDU.Frame.Data[0], 
               iOffset);

        memcpy ((void *)&pHeader, 
                (void *)&extractedPackets[nPackets].Data[0], 
                sizeof ( RawTM::XMM::PacketHeader ));

        Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    

        if ( (iCopied + iOffset) == (unsigned int)(pHeader.PacketLength + 7) ) {
          errstr << "NCTRS Fixer: Overspill processed" 
                 << Sas::message(Msg::App, Msg::Noisy);
          extractedPacketTimes[nPackets] = overspillTime;
          nPackets++;
        }
        else {
          errstr << "NCTRS Fixer: Overspill mismatch"
                 << Sas::warning("OverspillMismatch");
        }

      }
      else {
        errstr << "NCTRS Fixer: Discarding Overspill"
               << Sas::warning("OverspillDiscard");
      }
      
    }
    // END Packet Extraction 
    // BEGIN NORMAL PROCESSING
   lastVC = nctrsDU.Frame.FrameID.VCID;
    isOverspill = false;
    iCopied = 0;
    
    bool loop = true;
    bool cont = true;

    while ( loop && cont ) {
      
      iLeft = RawTM::NCTRS::data_size - iOffset;
      
      // Packet header split over frame boundary
      if ( sizeof(RawTM::XMM::PacketHeader) > iLeft ) {

        memcpy((void *)&overspill, 
               (void *)&nctrsDU.Frame.Data[ iOffset ],
               iLeft );

        overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

        loop = false;
        isOverspill = true;
        iCopied = iLeft;

        errstr << "NCTRS Fixer: Packet Header overspills" 
               << Sas::message(Msg::App, Msg::Noisy);

      }
      else {

        memcpy ((void *)&pHeader, 
                (void *)&nctrsDU.Frame.Data[ iOffset ],
                sizeof ( RawTM::XMM::PacketHeader ));

        Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    

        if ( (unsigned int)(pHeader.PacketLength + 7) > iLeft ) {
          
          memcpy((void *)&overspill, 
                 (void *)&nctrsDU.Frame.Data[ iOffset ],
                 iLeft );
          overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

          loop = false;
          isOverspill = true;
          iCopied = iLeft;

          errstr << "NCTRS Fixer: Packet overspills" 
                 << Sas::message(Msg::App, Msg::Noisy);
          
        }
        else {
          
          memcpy((void *)&extractedPackets[nPackets], 
                 (void *)&nctrsDU.Frame.Data[ iOffset], 
                 pHeader.PacketLength + 7);

          extractedPacketTimes[nPackets] = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);
          
          nPackets++;

          if (nPackets >= RawTM::NCTRS::max_packets) {
            errstr << "NCTRS Fixer: Too many packets"
                   << Sas::warning("maxPackets");
            cont = false;
          }
          iOffset += pHeader.PacketLength + 7;
          if (iOffset == RawTM::NCTRS::data_size) {
            loop = false;
          }
          errstr << "NCTRS Fixer: Packet processed" 
                 << Sas::message(Msg::App, Msg::Noisy);
        }

      } // END if packet lenght + 7 > iLeft
      
    } // END while loop and cont 

    // Call the user-defined process method
    if (dataProc->acceptFrame(nctrsDU.Header)) {
      dataProc->processFrame(nctrsDU);
    }
    
    errstr << nPackets << " packets extracted from Frame" 
           << Sas::message(Msg::App, Msg::Verbose);

   // CRx15
   if ( Filters->bFilterRequired )
    errstr << " Filtering by- APID " << Filters->FilterAPID
          << " Type : " << Filters->FilterType 
          << " SubType : " << Filters->FilterSubType 
          << Sas::message(Msg::App, Msg::Verbose);
   // END CRx15

   //////////////// PACKET DATA EXTRACTION 
   // Pass on the data
   for (unsigned int c = 0; c < nPackets; c++) {
    for (unsigned int i = 0 ; i < getMap().size(); i++) {
     PacketProcessor *p = (PacketProcessor *)getMap()[i];
     // CRx15
     //p->processPacket(&nctrsDU, extractedPackets[c], extractedPacketTimes[c]);
     // SPRx16
     //p->processPacket(Filters, extractedPackets[c], extractedPacketTimes[c]);
     p->processPacket(Filters,
                    iNumGeneric,iNumSpecific,MyarNoDFHTime, 
                    extractedPackets[c], extractedPacketTimes[c]);
    }
   }
   //////////////// END PACKET DATA EXTRACTION 

  // END this DU was on the right channel

//  }
//  else {
//    errstr << " Fixer: found wrong channel " << nctrsDU.Frame.FrameID.VCID 
//           << "  we only process " << channel 
//           << Sas::message(Msg::App, Msg::Verbose);
//  }

 // END of procData

 return;

}
// END CRx14

void NCTRSProcessor::extractData(RawTM::XMCS::DataUnit *DU){
  
  // Make a copy of the Raw DU data
  if ( DU->Size - RawTM::XMCS::header_size + sizeof(STime) != sizeof(RawTM::NCTRS::TMDU)) {
    
    errstr << "NCTRS Processor: XMM DU with NCTRS DU of wrong size!"
	   << Sas::warning("WrongSize");
    
    return;
    
  }    

  memcpy(&nctrsDU, DU->Data, DU->Size - RawTM::XMCS::header_size);
  
  // Byte swap NCTRS DU 
  Utils::sunByteSwap(nctrsDU.Header.rawPacketSize, sizeof(nctrsDU.Header.rawPacketSize));
  Utils::sunByteSwap(nctrsDU.Header.rawSCID, sizeof(nctrsDU.Header.rawSCID));
  Utils::sunByteSwap(nctrsDU.Header.rawGSID, sizeof(nctrsDU.Header.rawGSID));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Days, sizeof(nctrsDU.Header.rawTime.Days));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Millis, sizeof(nctrsDU.Header.rawTime.Millis));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Micros, sizeof(nctrsDU.Header.rawTime.Micros));
  Utils::sunByteSwap(nctrsDU.Header.rawQuality, sizeof(nctrsDU.Header.rawQuality));

  errstr << "NCTRS Processor: NCTRS Header: " 
	 << " FrameSize " << nctrsDU.Header.PacketSize 
	 << " SCID " << nctrsDU.Header.SCID 
	 << " DSType " << (int)nctrsDU.Header.DSType 
	 << " VCID " << (int)nctrsDU.Header.VCID 
	 << " GSID " << nctrsDU.Header.GSID 
	 << " Quality " << nctrsDU.Header.Quality 
	 << " Time " 
	 << nctrsDU.Header.Time.Days << ":"
	 << nctrsDU.Header.Time.Milliseconds << ":"
	 << nctrsDU.Header.Time.Microseconds             
	 << Sas::message(Msg::App, Msg::Verbose);

  if (nctrsDU.Header.Quality == RawTM::NCTRS::bad) {

    errstr << "NCTRS Processor: Bad quality NCTRS frame"
	   << Sas::warning("BadFrame");
    
    return;
    
  }

  if ( nctrsDU.Header.SCID != 193 ) {
    
    errstr << "NCTRS Processor: Data not from S/C: S/C ID = " << nctrsDU.Header.SCID
	   << Sas::warning("BadSC");
    
    return;

  }

  // NEW CRx20 bti filtering per g/s
  // It may be used the fake gsid instead of generic boolean
  // but in that casee, a loop over bti should be run every NCTRS unit (slow)

  STime nctrsDUTime = Utils::cds2STime(nctrsDU.Header.Time); 
  //errstr << "NCTRS Processor: Checking if apply BTI:" 
  //       << bUseGsBti << Sas::message(Msg::App, Msg::Sparse);
  if ( bFullBTI ) {
       std::vector<int16>::const_iterator gsIDiter = bti_gsid_Filter.begin();
       for (std::vector<BTI_pair>::const_iterator timeIter = timeFilter.begin();
       timeIter!= timeFilter.end();
       timeIter++, gsIDiter++ ) {
  
       //errstr << "NCTRS Processor: Comparing:" 
       //  << nctrsDUTime << nctrsDU.Header.GSID
       //  << Sas::message(Msg::App, Msg::Sparse);

          if ( (nctrsDUTime > (*timeIter).first) && 
               (nctrsDUTime < (*timeIter).second) &&
               (nctrsDU.Header.GSID == (*gsIDiter))
             ) {
               errstr << "NCTRS Processor: WARNING GS ID found in BTI file " 
               << nctrsDUTime << nctrsDU.Header.GSID
               << Sas::warning("IgnoredGS");
               return;
               }

        } // End for loop over bti rows 

  } // End if bti filter should be used 

  // END CRx20

  if ( (nctrsDU.Header.GSID != RawTM::NCTRS::VIL1 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::REDU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::VIL2 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::KRU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::PER ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::NNO ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::AGO ) &&
       // NEW CRx19
       (nctrsDU.Header.GSID != RawTM::NCTRS::KGAL ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::DON ) ) {

    errstr << "NCTRS Processor: WARNING Bad GS ID" << nctrsDU.Header.GSID
	   << Sas::warning("BadGS");
    
    return;

  }
  
  // Byte swap Transfer Frame
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameID, sizeof(nctrsDU.Frame.rawFrameID));
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameStatus, sizeof(nctrsDU.Frame.rawFrameStatus));
  Utils::sunByteSwap(nctrsDU.Frame.rawCLCW, sizeof(nctrsDU.Frame.rawCLCW));

  errstr << "NCTRS Processor: NCTRS Transfer Frame: " 
	 << " OpControl " << nctrsDU.Frame.FrameID.OpControl     
	 << " VCID " << nctrsDU.Frame.FrameID.VCID	  
	 << " SCID " << nctrsDU.Frame.FrameID.SCID          
	 << " Version " << nctrsDU.Frame.FrameID.Version	  
	 << " MFC " << (int)nctrsDU.Frame.MFC           
	 << " VFC " << (int)nctrsDU.Frame.VFC           
	 << " FirstHeader " << nctrsDU.Frame.FrameStatus.FirstHeader   
	 << " SegID " << nctrsDU.Frame.FrameStatus.SegID	 
	 << " PacketOrder " << nctrsDU.Frame.FrameStatus.PacketOrder   
  	 << " SyncFlag " << nctrsDU.Frame.FrameStatus.SyncFlag      
	 << " SecHeaderFlag " << nctrsDU.Frame.FrameStatus.SecHeaderFlag 
	 << " CLCW " << hex << nctrsDU.Frame.CLCW << dec 
   	 << Sas::message(Msg::App, Msg::Verbose);

  // Check for extraneous idle frames (none should ever get here...)
  if (nctrsDU.Frame.FrameStatus.FirstHeader == 0x07FE) {
    errstr << "NCTRS Processor: Idle TM Trame found on VC" << (int)nctrsDU.Header.VCID 
           // SPRx32 replaced the message warning
	   //<< Sas::warning("duDataGap");
	   << Sas::warning("duIdleFrame");
  }

  // Correct Frame Time
  nctrsDU.CorrectedFrameTime = correctFrameTime(nctrsDU.Header.Time, nctrsDU.Header.GSID);

  // If this DU is on the right channel
  if (nctrsDU.Frame.FrameID.VCID == channel) {

    RawTM::XMM::PacketHeader pHeader;
    unsigned int iLeft;
    
    unsigned int iOffset = nctrsDU.Frame.FrameStatus.FirstHeader;
    nPackets = 0;
    
    // Packet Extraction goes here
    if (isFirstFrame) {

      errstr << "NCTRS Processor: Processing first frame" 
	     << Sas::message(Msg::App, Msg::Noisy);

      
      isFirstFrame = false;
      
    }
    else if (isOverspill) {

      errstr << "NCTRS Processor: Overspill to be processed" 
	     << Sas::message(Msg::App, Msg::Noisy);
      
      if (nctrsDU.Frame.FrameID.VCID != lastVC) {
	
	errstr << "NCTRS Processor: Frame from wrong VC!"
	       << Sas::warning("WrongVC");
	
      }
      else if ( (iOffset + iCopied) <= sizeof( RawTM::XMM::Packet ) ) {
	
	// Copy the overspill into the new packet
	extractedPackets[nPackets] = overspill;
	
	// iOffset gives the number of bytes not transmitted in the packet from the last frame
	memcpy((void *)&extractedPackets[nPackets].Data[iCopied], 
	       (void *)&nctrsDU.Frame.Data[0], 
	       iOffset);
	
	memcpy ((void *)&pHeader, 
		(void *)&extractedPackets[nPackets].Data[0], 
		sizeof ( RawTM::XMM::PacketHeader ));
	
	Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    
	
	if ( (iCopied + iOffset) == (unsigned int)(pHeader.PacketLength + 7) ) {

	  errstr << "NCTRS Processor: Overspill processed" 
		 << Sas::message(Msg::App, Msg::Noisy);

	  extractedPacketTimes[nPackets] = overspillTime;
	  nPackets++;
	  
	}
	else {
	  
	  errstr << "NCTRS Processor: Overspill mismatch"
		 << Sas::warning("OverspillMismatch");
	  
	}
	
      }
      else {
	
	errstr << "NCTRS Processor: Discarding Overspill"
	       << Sas::warning("OverspillDiscard");
	
      }
      
    }
    
    // Setup for normal processing
    lastVC = nctrsDU.Frame.FrameID.VCID;
    isOverspill = false;
    iCopied = 0;
    
    bool loop = true;
    bool cont = true;

    while ( loop && cont ) {
      
      iLeft = RawTM::NCTRS::data_size - iOffset;
      
      // Packet header split over frame boundary
      if ( sizeof(RawTM::XMM::PacketHeader) > iLeft ) {
	
	memcpy((void *)&overspill, 
	       (void *)&nctrsDU.Frame.Data[ iOffset ],
	       iLeft );
	
	overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

	loop = false;
	isOverspill = true;
	iCopied = iLeft;
	
	errstr << "NCTRS Processor: Packet Header overspills" 
	       << Sas::message(Msg::App, Msg::Noisy);
	
      }
      else {
	
	memcpy ((void *)&pHeader, 
		(void *)&nctrsDU.Frame.Data[ iOffset ],
		sizeof ( RawTM::XMM::PacketHeader ));
	
	Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    
	
	if ( (unsigned int)(pHeader.PacketLength + 7) > iLeft ) {
	  
	  memcpy((void *)&overspill, 
		 (void *)&nctrsDU.Frame.Data[ iOffset ],
		 iLeft );
	  
	  overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

	  loop = false;
	  isOverspill = true;
	  iCopied = iLeft;

	  errstr << "NCTRS Processor: Packet overspills" 
		 << Sas::message(Msg::App, Msg::Noisy);
	  
	}
	else {
	  
	  memcpy((void *)&extractedPackets[nPackets], 
		 (void *)&nctrsDU.Frame.Data[ iOffset], 
		 pHeader.PacketLength + 7);

	  extractedPacketTimes[nPackets] = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);
	  
	  nPackets++;
	  
	  if (nPackets >= RawTM::NCTRS::max_packets) {
	    errstr << "NCTRS Processor: Too many packets"
		   << Sas::warning("maxPackets");
	    
	    cont = false;
	  }
	  
	  iOffset += pHeader.PacketLength + 7;
	  
	  if (iOffset == RawTM::NCTRS::data_size) {
	    loop = false;
	  }

	  errstr << "NCTRS Processor: Packet processed" 
		 << Sas::message(Msg::App, Msg::Noisy);
	  
	}
	
      }
      
    }

    // Call the user-defined process method
    if (dataProc->acceptFrame(nctrsDU.Header)) {
      dataProc->processFrame(nctrsDU);
    }
    
    errstr << nPackets << " packets extracted from Frame" 
	   << Sas::message(Msg::App, Msg::Noisy);
   
    // NEw CRx15 fix filters, even will not be used here
    FilterInfo Filters; 
    Filters.bFilterRequired=false;
    // ENd CRx15 

    // NEW SPRx16 DFH Info, even when not really used here
    VTMZZ_TNODFHTIME MyarNoDFHTime[1];
    // END SPRx16
 
    // Pass on the data
    for (unsigned int c = 0; c < nPackets; c++) {
      
      for (unsigned int i = 0 ; i < getMap().size(); i++) {
	
	PacketProcessor *p = (PacketProcessor *)getMap()[i];

        // CRx15
	//p->processPacket(&nctrsDU, extractedPackets[c], extractedPacketTimes[c]);
        // SPRx16
	//p->processPacket(&Filters, extractedPackets[c], extractedPacketTimes[c]);
        p->processPacket(&Filters,
                    0,0,MyarNoDFHTime, 
                    extractedPackets[c], extractedPacketTimes[c]);
	
      }
      
    }
    
  }
      
}


STime NCTRSProcessor::correctFrameTime(Utils::CDSTime Time, unsigned short GSID) {

  STime uncorrectedXMCSFrameTime = Utils::cds2STime(Time);

  double xmcsOBDHDelay = delayHandler.getXMCSOBDHdelay(uncorrectedXMCSFrameTime);
  double xmcsGSDelay   = delayHandler.getXMCSGSdelay(uncorrectedXMCSFrameTime, GSID);
  
  double actualOBDHDelay = delayHandler.getActualOBDHdelay(uncorrectedXMCSFrameTime);
  double actualGSDelay   = delayHandler.getActualGSdelay(uncorrectedXMCSFrameTime, GSID);
  
  // Before 2002-01-22T12:00:00Z, the XMCS overwrote the ERT. Correct for this
  // by reversing the changes
  
  // Frame time, corrected for XMCS overwrite
  STime correctedXMCSFrameTime;
  double signalDelay;
  
  // Calculate signal delay from ERT + XMCS correction (since the XMCS would have 
  // subtracted the GS Delay from the ERT
  if (uncorrectedXMCSFrameTime < refXMCSCorrection) {
    
    STime PropDelayCorrected = uncorrectedXMCSFrameTime + xmcsGSDelay;
    signalDelay = delayHandler.getSignalDelay(PropDelayCorrected, GSID);
    
    correctedXMCSFrameTime = uncorrectedXMCSFrameTime + signalDelay;

    correctedXMCSFrameTime = correctedXMCSFrameTime + xmcsGSDelay;

    correctedXMCSFrameTime = correctedXMCSFrameTime + xmcsOBDHDelay;

  }
  else {
    correctedXMCSFrameTime = uncorrectedXMCSFrameTime;
    signalDelay = delayHandler.getSignalDelay(correctedXMCSFrameTime, GSID);	
  }

  double signalDelay2 = delayHandler.getSignalDelay(correctedXMCSFrameTime - signalDelay, GSID);
    
  // Now add the various delays
  STime correctedFrameTime = correctedXMCSFrameTime;
  
  correctedFrameTime = correctedFrameTime - signalDelay2;
  
  correctedFrameTime = correctedFrameTime - actualGSDelay;

  correctedFrameTime = correctedFrameTime - actualOBDHDelay;

  return correctedFrameTime;

}


STime NCTRSProcessor::calculatePacketTime(unsigned int iOffset, STime frameTime) {

  STime pktTime = frameTime + ((iOffset * 8) / 70000.0);

  return pktTime;

}

// NEW CRx14
void NCprintHex(unsigned char *buf, int len) {
  int iBloqs=len/4;
  int iRest=len%4;
  int iBase;

  cout << "Packet Lenght is " << len << endl;
  cout << "Packet Bloqs are " << iBloqs << endl;
  cout << "Packet Rest is " << iRest << endl;
  cout << hex;
  for (int i = 0; i < iBloqs; i++) {
    iBase=i*4;
    cout << setw(2) << setfill('0') << (int)buf[iBase];
    cout << setw(2) << setfill('0') << (int)buf[iBase+1];
    cout << setw(2) << setfill('0') << (int)buf[iBase+2];
    cout << setw(2) << setfill('0') << (int)buf[iBase+3];
    cout << "  ";

    if ( ((i+1)%8) == 0) cout << " " << endl;

  }
  for (int i = 0; i < iRest; i++) {
    iBase=iBloqs*4;
    cout << hex << setw(2) << setfill('0') << (int)buf[i+iBase];
  } 
  cout << endl;
  cout << dec;
  // END CRx10

}


