// CRx14                Infrastructure for dump/fixer 
// CRx15                Applied filter capabilities
// SPRx16 jcv 20-May-09 DHF guessing capabilities
// CRx20  jcv 21-Sep-10 BTI gsid to be ignored if within given timespan

#ifndef NCTRSPROCESSOR_H
#define NCTRSPROCESSOR_H

#include "DUProcessor.h"
#include "DataDefinition.h"
#include "Utils.h"
#include "FrameConsumer.h"

#include "Delay.h"

// NEw CRx15
#include "FilterDefinition.h"
// END CRx15

// NEW SPRx16
#include "dfhtimestamp.h"
// END SPRx16

// NEW CRx20
#include "Dal.h"
#include "CalOalUtility.h"
typedef std::pair<STime,STime> BTI_pair;
// END CRx20

class NCTRSProcessor : public DUProcessor {

public:

  enum VC {VC0 = 0, VC7 = 7};
  
  RawTM::NCTRS::TMDU     nctrsDU;

  void                  extractData(RawTM::XMCS::DataUnit *DU);
  // NEW CRx14
  // NEW CRx15
  //void   procData(RawTM::XMCS::DataUnit *DU);
  // NEW SPRx16
  //void   procData(RawTM::XMCS::DataUnit *DU, FilterInfo *Filters);
  void     procData(RawTM::XMCS::DataUnit *DU, 
                    FilterInfo *Filters,
                    int iNumGeneric, int iNumSpecific,
                    VTMZZ_TNODFHTIME *MyarNoDFHTime);

  DUProcessor::type     processorType() { return DUProcessor::TM; }
  
                        NCTRSProcessor(FrameConsumer *proc, NCTRSProcessor::VC c);
                        ~NCTRSProcessor() { }
  
 private:

  STime              correctFrameTime(Utils::CDSTime Time, unsigned short GSID);
  STime              calculatePacketTime(unsigned int iOffset, STime frameTime);
  FrameConsumer     *dataProc;
  NCTRSProcessor::VC channel;

  Delay              delayHandler;

  bool               isFirstFrame;
  bool               isOverspill;
  unsigned int       lastVC;
  unsigned int       iCopied;
  RawTM::XMM::Packet overspill;
  RawTM::XMM::Packet extractedPackets[ RawTM::NCTRS::max_packets ];
  STime              extractedPacketTimes[ RawTM::NCTRS::max_packets ];
  STime              overspillTime;  
  unsigned int       nPackets;

  // Reference point for XMCS correction
  STime refXMCSCorrection;

  // NEW CRx20
  std::vector<BTI_pair> timeFilter;
  std::vector<int16> bti_gsid_Filter;
  bool bFullBTI;


};

#endif
