// CRx10  jcv 18-Mar-09 Added functionality for HEX tmcheck information
// CRx11  jcv 31-Mar-09 Fixed verbosity output 
// SPRx14 jcv 21-Apr-09 Fixed dump capabilities
// CRx15  jcv 05-May-09 Filter capabilities
// SPRx16 jcv 20-May-09 DFH Guessing capabilities

#include "PacketProcessor.h"

#include "errstr.h"

void printHex(unsigned char *buf, int len);

// CRx15
void fprintHex(FilterInfo *Filters, unsigned char *buf, int len);

// SPRx16 DFH guessing capabilities
int  fiBinarySearch(
        VTMZZ_TNODFHTIME * aiData, /* I: Sorted array to be searched.      */
        int iStartPos,             /* I: Start position in array.          */
        int iNumEntries,           /* I: Number of entries to be searched. */
        long int iValue );         /* I: Value to be searched for.         */


PacketProcessor::PacketProcessor(PacketConsumer *proc) : dataProc(proc) {

  haveSTSP = false;
  
}

// CRx15
//void PacketProcessor::processPacket(RawTM::NCTRS::TMDU *frame, RawTM::XMM::Packet &pkt, STime &time) {
// SPRx16
//void PacketProcessor::processPacket(FilterInfo *Filters, RawTM::XMM::Packet &pkt, STime &time) {
void PacketProcessor::processPacket(FilterInfo *Filters,
                             int iNumGeneric, int iNumSpecific,
                             VTMZZ_TNODFHTIME *MyarNoDFHTime,
                             RawTM::XMM::Packet &pkt, STime &time) {

  // SPRx16 Time in DFH guessing capabilities
  bool bHasTime=false;
  // END SPRx16

  xmmPkt = pkt;
  // CRx15
  //xmmFrame = frame;

  //if ( Filters->bFilterRequired ) {
  //  errstr << " Packet Filtering by- APID " << Filters->FilterAPID
  //        << " Type : " << Filters->FilterType 
  //        << " SubType : " << Filters->FilterSubType 
  //        << " SID : " << Filters->FilterSID 
  //        << Sas::message(Msg::App, Msg::Verbose);
  //}
  // END CRx15

  // NEW CRx10 (TBchecked the performance)
  errstr << "PACKET DETAILS BEGINS " << Sas::message(Msg::App, Msg::Noisy);
  errstr << "====================" << Sas::message(Msg::App, Msg::Noisy);
  //if(errHandler.verbosity(Msg::App) > Msg::Noisy )
  //    printHex(pkt.Header, sizeof( RawTM::XMM::PacketHeader ));

  // END CRx10

  Utils::sunByteSwap(xmmPkt.Header.rawPacketID, sizeof(xmmPkt.Header.rawPacketID));    
  Utils::sunByteSwap(xmmPkt.Header.rawPacketLength, sizeof(xmmPkt.Header.rawPacketLength));    

  errstr << "Packet Time: " 
	 << time.toString(6)
   	 << Sas::message(Msg::App, Msg::Noisy);

  errstr << "Packet Header: " 
	 << " Version " << xmmPkt.Header.PacketID.VersionNumber
	 << " Type " << xmmPkt.Header.PacketID.Type
    	 << " DFH Flag " << xmmPkt.Header.PacketID.DataFieldHeaderFlag
	 << " APID " << xmmPkt.Header.PacketID.APID
	 << " Sequence Count " << xmmPkt.Header.PacketID.SequenceCount
	 << " Segmentation Flag " << xmmPkt.Header.PacketID.SeqmentationFlag
	 << " Length " << xmmPkt.Header.PacketLength
   	 << Sas::message(Msg::App, Msg::Noisy);

  if (xmmPkt.Header.PacketID.APID == 0) {
    // Time Packet
    
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s1, 2);  
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s2, 2);  
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s3, 2);  

    // Have to perform this monstrosity due to gcc's insistence on padding
    // structures, regardless of directive i.e. '__attribute ((packed))'
    // Not a disaster, since we are working with a copy of the packet only and
    // there is no data field in an STPS which would be overwritten at bytes 0 and 1.
    unsigned short s = xmmPkt.DFHeader.RawTime2.s3;
    xmmPkt.DFHeader.Time.Coarse = xmmPkt.DFHeader.RawTime2.s2 | (xmmPkt.DFHeader.RawTime2.s1 << 16);
    xmmPkt.DFHeader.Time.Fine = s;

    errstr << "Time: " << hex
	   << " S-Field " << (int)xmmPkt.DFHeader.Time.SField
	   << " P-Field " << (int)xmmPkt.DFHeader.Time.PField 
	   << " Coarse " << (int)xmmPkt.DFHeader.Time.Coarse
	   << " Fine " << (int)xmmPkt.DFHeader.Time.Fine 
	   << dec
	   << Sas::message(Msg::App, Msg::Noisy);

    lastSTSP = (int)xmmPkt.DFHeader.Time.Coarse;
    haveSTSP = true;

  }
  else if (xmmPkt.Header.PacketID.APID == 0x7ff) {
    errstr << "IDLE"
           << Sas::message(Msg::App, Msg::Noisy);
    // idle
  }
  else {
    // TM

    Utils::sunByteSwap(xmmPkt.DFHeader.rawTM, sizeof(xmmPkt.DFHeader.rawTM));

    errstr << "Data Field: " 
	   << " Type " << xmmPkt.DFHeader.TM.PacketType
	   << " SubType " << xmmPkt.DFHeader.TM.PacketSubtype
	   << " Checksum " << xmmPkt.DFHeader.TM.ChecksumFlag
	   << " Spare " << xmmPkt.DFHeader.TM.Spare
	   << Sas::message(Msg::App, Msg::Noisy);    

    // NEW SPRx16 DFH guessing capabilities
    //if( hasTimeinDFH(xmmPkt.Header.PacketID.APID, xmmPkt.DFHeader.TM.PacketType, xmmPkt.DFHeader.TM.PacketSubtype) ) {

    bHasTime=hasTimeinDFH(xmmPkt.Header.PacketID.APID, xmmPkt.DFHeader.TM.PacketType, xmmPkt.DFHeader.TM.PacketSubtype,iNumGeneric,iNumSpecific,MyarNoDFHTime);

    if( bHasTime ) {

      Utils::sunByteSwap(xmmPkt.DFHeader.rawTMTime.rawFine, sizeof(xmmPkt.DFHeader.rawTMTime.rawFine));
      Utils::sunByteSwap(xmmPkt.DFHeader.rawTMTime.rawCoarse, sizeof(xmmPkt.DFHeader.rawTMTime.rawCoarse));

      // ACC OBT times are missing the top 8msb, copy from last STSP.
      if (xmmPkt.Header.PacketID.APID == 640) {
	if(haveSTSP) {
	  xmmPkt.DFHeader.TMTime.CoarseTime = (xmmPkt.DFHeader.TMTime.CoarseTime & 0x00ffffff) | (lastSTSP & 0xff000000);
	}
	else {
	  errstr << "No STSP available to correct APID 640 OBT"
		 << Sas::warning("cannotCorrect640OBT");
	}
      }

      errstr << "Data Field Time: " 
	     << hex
	     << " Coarse Time " << xmmPkt.DFHeader.TMTime.CoarseTime
	     << " Fine Time " << xmmPkt.DFHeader.TMTime.FineTime
	     << dec
	     << Sas::message(Msg::App, Msg::Noisy);

    }

  }

  // CRx15
  // SPRx16 Enahnced with Time in DFH guessing
  // SID treatment
  unsigned char rawSID;
  if (bHasTime)
     rawSID=pkt.Data[14];
  else
     rawSID=pkt.Data[8];

  errstr << "SID: " 
    << setw(2) << setfill('0') << (int)rawSID
    << Sas::message(Msg::App, Msg::Noisy);

  if (xmmPkt.Header.PacketID.APID != 0) {
     errstr << "Packet ID Info: " 
	 << " APID " << xmmPkt.Header.PacketID.APID
	 << " Type " << xmmPkt.DFHeader.TM.PacketType
	 << " SubType " << xmmPkt.DFHeader.TM.PacketSubtype
         << " SID: " << setw(2) << setfill('0') << (unsigned int)rawSID
         << Sas::message(Msg::App, Msg::Noisy);
  }
  // END CRx15

  // CRx15
  if ( Filters->bFilterRequired ) {
     if ( (xmmPkt.Header.PacketID.APID == Filters->FilterAPID) &&
          (xmmPkt.DFHeader.TM.PacketType == Filters->FilterType) &&
          ((unsigned int)rawSID == Filters->FilterSID) && 
          (xmmPkt.DFHeader.TM.PacketSubtype == Filters->FilterSubType) ) 
          {

          // Dump the timestamp
          Filters->ofsPacketFile << time.toString(6) << " ";

          // Dump the selected (or all ) packet bytes
          fprintHex(Filters, pkt.Data, 7+xmmPkt.Header.PacketLength);

          // Additional adhoc treatment may be inserted here
          //unsigned short ExpId;
          //Filters->ofsPacketFile << sizeof(ExpId) << " ";
          //ExpId=(unsigned int)pkt.Data[20]*256+(unsigned int)pkt.Data[21];
          //Filters->ofsPacketFile << ExpId << " " << endl;

          }
  } else 
         if(errHandler.verbosity(Msg::App) > Msg::Noisy ) 
                 printHex(pkt.Data, 7+xmmPkt.Header.PacketLength);

  // SPRx14 At some stage, for just dumping, no stsp is needed,
  // may be checked and disabled if needed
  // Here calls to PacketConsumer processor
  if (dataProc->acceptPacket(xmmPkt.Header, xmmPkt.DFHeader)) {
    dataProc->processPacket(xmmPkt, time);
  }
  
}

// NEW SPRx16 New required function
// CRH heritage
// This function carries out a binary search on a ascendingly sorted signed
// integer array and returns the number of the array element whose value
// matches that specified. If the value is not found -1 is returned.

int  fiBinarySearch(
        VTMZZ_TNODFHTIME * aiData, /* I: Sorted array to be searched.      */
        int iStartPos,             /* I: Start position in array.          */
        int iNumEntries,           /* I: Number of entries to be searched. */
        long int iValue          /* I: Value to be searched for.         */
        )
{

/*+DATA************************************/

   int      iLow  = iStartPos;
   int      iHigh = iStartPos + iNumEntries;
   int      iAvg;

/*-DATA************************************/

   /*+ Search the array. +*/

   for( ; iHigh >= iLow; )
   {
      iAvg = (iLow + iHigh)/2;

      if( iValue < aiData[ iAvg ].uMap.iPktID )
      {
         iHigh = iAvg - 1;
      }
      else if( iValue > aiData [ iAvg ].uMap.iPktID )
      {
         iLow = iAvg + 1;
      }
      else
      {
         return iAvg;
      }
   }

   return -1;

}
// END SPRx16

// NEW x16 DFH guessing facilities activated, full refactoring
//  of the previously existing dummy fucntion
//bool PacketProcessor::hasTimeinDFH(unsigned int APID, unsigned short PacketType, unsigned short PacketSubtype) {

// Some comments from CH heritage here
//This module checks if a packet specified by type, subtype and APID has got
//a timestamp in the data field header.

// REMEMBER!!!!
// The config file contain a list of all Packets that do NOT contain time
// i.e. if found in the list, hasTime will return false
// i.e. if not found in the list, hasTime will return true 

//Two special cases should be noted these are;

// NOTE 1: APID = 000 (hex), this only applies to STSP packets, hence if an
//         APID of 000 (hex) is passed to this function it will always return
//         FALSE (irrespective of type and subtype specified) as the STSP does
//         not have a timestamp in the DFH (or even a data field header).

// NOTE 2: APID = 7FF (hex), this only applies to idle packets, hence if an
//         APID of 7FF (hex) is passed to this function it will always return
//         FALSE (irrespective of type and subtype specified) as the idle does
//        not have a timestamp in the DFH (or even a data field header).

bool PacketProcessor::hasTimeinDFH(unsigned int APID, 
                            unsigned short PacketType, 
                            unsigned short PacketSubtype,
                            int iNumSpecific, int iNumGeneric,
                             VTMZZ_TNODFHTIME *MyarNoDFHTime) {

  int               iIndex;
  VTMZZ_TNODFHTIME  rSearchValue;

  errstr << "Is there time in DFH for APID " << APID 
         << " Type " << PacketType 
         << " Subtype " << PacketSubtype 
         << " ... " << Sas::message(Msg::App, Msg::Noisy);

   /* Check if STSP or idle packet and if so return -1 as these do not */
   /* have timestamps in the DFH, or indeed even have a DFH. STSP      */
   /* packets have an APID of 000 (hex) and IDLE packets an APID of    */
   /* 7FF (hex). Thus if an APID of either of these values is          */
   /* specified return FALSE irrespective of the packet type and       */
   /* subtype specified.                                               */

   if( (APID == 0x000) || (APID == 0x7FF) )
      {
      errstr << "...STSP or Idle, anyhow no time in DFH"
        << Sas::message(Msg::App, Msg::Noisy);
      return false;
      }

   /* First search generic part of list, create required search value; */
   /* NOTE: APID of -1 is used to designate generic value.             */

   rSearchValue.uMap.rPktID.cType    = PacketType;
   rSearchValue.uMap.rPktID.cSubtype = PacketSubtype;
   rSearchValue.uMap.rPktID.siAPID   = -1;

   iIndex = fiBinarySearch( MyarNoDFHTime,
                            0,
                            iNumGeneric,
                            rSearchValue.uMap.iPktID );

   /* If not found in generic list then check specific list.           */

   if( iIndex == -1 )
   {
      /* Search Specifc list, update key.                              */

      rSearchValue.uMap.rPktID.siAPID = APID;

      iIndex = fiBinarySearch( MyarNoDFHTime,
                               iNumGeneric,
                               iNumSpecific,
                               rSearchValue.uMap.iPktID );
   }

   // Return TRUE if not found in file : timestamp in data field header
   if( iIndex == -1 ) {
      errstr << "...Timestamp in DFH " << Sas::message(Msg::App, Msg::Noisy);
      return true;
    }

   // Okay if we've got here, it was found, so no timestamp in DFH 
   errstr << "...NO Timestamp in DFH " << Sas::message(Msg::App, Msg::Noisy);
   return false;

}

void printHex(unsigned char *buf, int len) {
  // NEw CRx10
  int iBloqs=len/4;
  int iRest=len%4;
  int iBase;
  // END CRx10

  // SPRx14 finally commented out
  //for (int i = 0; i < len; i++) {
  //  cout << hex << setw(2) << setfill('0') << (int)buf[i];
  //}
  //cout << endl;
  //cout << dec;

  // NEW CRx10
  //cout << "Packet Lenght is " << len 
  // << " Packet Bloqs are " << iBloqs 
  // << " Packet Rest is " << iRest << endl;
  cout << hex;

  for (int i = 0; i < iBloqs; i++) {
    iBase=i*4;
    cout << setw(2) << setfill('0') << (int)buf[iBase];
    cout << setw(2) << setfill('0') << (int)buf[iBase+1];
    cout << setw(2) << setfill('0') << (int)buf[iBase+2];
    cout << setw(2) << setfill('0') << (int)buf[iBase+3];
    cout << "  ";

    if ( ((i+1)%8) == 0) cout << " " << endl;

  }
  for (int i = 0; i < iRest; i++) {
    iBase=iBloqs*4;
    cout << hex << setw(2) << setfill('0') << (int)buf[i+iBase];
  } 
  cout << endl;
  cout << dec;
  // END CRx10

}

// CRx15
void fprintHex(FilterInfo *Filters, unsigned char *buf, int len) {
  int iBloqs=len/4;
  int iRest=len%4;
  int iBase;

  // first checks the mask (0 do not dump, 1 will dump)
  // default is all 1's :do full dump 
  // IMPORTANT: when string is finished, assumed 0's: do not dump unless
  //             requested
  // BUT If there is no 0 in mask, then all are 1: means dump everything
  //     even further than mask lenght

  // If there is no 0 in mask, then all are 1: dump everything
  if (Filters->FilterMask.find("0") == string::npos) 
  {

  Filters->ofsPacketFile << hex;

  for (int i = 0; i < iBloqs; i++) {
    iBase=i*4;
    Filters->ofsPacketFile << setw(2) << setfill('0') << (int)buf[iBase]
     << setw(2) << setfill('0') << (int)buf[iBase+1]
     << setw(2) << setfill('0') << (int)buf[iBase+2]
     << setw(2) << setfill('0') << (int)buf[iBase+3]
     << "  ";

    //if ( ((i+1)%8) == 0) Filters->ofsPacketFile << " " << endl;

  }
  for (int i = 0; i < iRest; i++) {
    iBase=iBloqs*4;
    Filters->ofsPacketFile << hex 
          << setw(2) << setfill('0') << (int)buf[i+iBase];
  } 
  Filters->ofsPacketFile << dec << endl;

  //Filters->ofsPacketFile.flush();
  } 

  /// End Dump all //////////////////////////////////////////

  else
  // There was at least one 0 in mask, we will dump masking the filter
  {
    Filters->ofsPacketFile << hex;

    for (int i = 0; i < len ; i++) {
       int iLen=(Filters->FilterMask.size());
       if (i < iLen )
          // Dump only if 1 is present
          {
          char cMask=Filters->FilterMask.at(i);
          if ( cMask == '1' ) 
             Filters->ofsPacketFile << setw(2) << setfill('0') << (int)buf[i]; 
          } 
    } // End for the mask string loop

    Filters->ofsPacketFile << dec << endl;

  }

  // End dumping

}

