// CRx18 jcv 07-Jul-09 Original 
// CRx18b jcv 08-Jul-09 Refinements for SID filtering

#include "TMPacketProcessor.h"

#include "errstr.h"

TMPacketProcessor::TMPacketProcessor(PacketConsumer *proc) : dataProc(proc) {

  // ... 
  bDummie = true;
  
}

void TMPacketProcessor::processPacket(RawTM::XMM::Packet &pkt, STime &time) {

  bool bHasTime=false;

  xmmPkt = pkt;

  Utils::sunByteSwap(xmmPkt.Header.rawPacketID, sizeof(xmmPkt.Header.rawPacketID));    
  Utils::sunByteSwap(xmmPkt.Header.rawPacketLength, sizeof(xmmPkt.Header.rawPacketLength));    

  errstr << "PACKET DETAILS BEGINS " << Sas::message(Msg::App, Msg::Noisy);
  errstr << "====================" << Sas::message(Msg::App, Msg::Noisy);

  errstr << "Ground Packet Time (as calculated from offset and Frame time): " 
	 << time.toString(6)
   	 << Sas::message(Msg::App, Msg::Noisy);

  errstr << "Packet Header: " 
	 << " Version " << xmmPkt.Header.PacketID.VersionNumber
	 << " Type " << xmmPkt.Header.PacketID.Type
    	 << " DFH Flag " << xmmPkt.Header.PacketID.DataFieldHeaderFlag
	 << " APID " << xmmPkt.Header.PacketID.APID
	 << " Sequence Count " << xmmPkt.Header.PacketID.SequenceCount
	 << " Segmentation Flag " << xmmPkt.Header.PacketID.SeqmentationFlag
	 << " Length " << xmmPkt.Header.PacketLength
   	 << Sas::message(Msg::App, Msg::Noisy);

  if (xmmPkt.Header.PacketID.APID == 0) {
    // Time Packet
    
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s1, 2);  
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s2, 2);  
    Utils::sunByteSwap((unsigned char *)&xmmPkt.DFHeader.RawTime2.s3, 2);  

    // Have to perform this monstrosity due to gcc's insistence on padding
    // structures, regardless of directive i.e. '__attribute ((packed))'
    // Not a disaster, since we are working with a copy of the packet only and
    // there is no data field in an STPS which would be overwritten at bytes 0 and 1.
    unsigned short s = xmmPkt.DFHeader.RawTime2.s3;
    xmmPkt.DFHeader.Time.Coarse = xmmPkt.DFHeader.RawTime2.s2 | (xmmPkt.DFHeader.RawTime2.s1 << 16);
    xmmPkt.DFHeader.Time.Fine = s;

    errstr << "Time: " << hex
	   << " S-Field " << (int)xmmPkt.DFHeader.Time.SField
	   << " P-Field " << (int)xmmPkt.DFHeader.Time.PField 
	   << " Coarse " << (int)xmmPkt.DFHeader.Time.Coarse
	   << " Fine " << (int)xmmPkt.DFHeader.Time.Fine 
	   << dec
	   << Sas::message(Msg::App, Msg::Noisy);

  }
  else if (xmmPkt.Header.PacketID.APID == 0x7ff) {
    errstr << "IDLE"
           << Sas::message(Msg::App, Msg::Noisy);
    // idle
  }
  else {
    // TM

    Utils::sunByteSwap(xmmPkt.DFHeader.rawTM, sizeof(xmmPkt.DFHeader.rawTM));

    errstr << "Data Field: " 
	   << " Type " << xmmPkt.DFHeader.TM.PacketType
	   << " SubType " << xmmPkt.DFHeader.TM.PacketSubtype
	   << " Checksum " << xmmPkt.DFHeader.TM.ChecksumFlag
	   << " Spare " << xmmPkt.DFHeader.TM.Spare
	   << Sas::message(Msg::App, Msg::Noisy);    

    // XXXXXXXXXXXx
    // Has time in the DFH????

    bHasTime = false;
    if (dataProc->hasTimePacket(xmmPkt) ) bHasTime = true;
 
    if( bHasTime ) {

      Utils::sunByteSwap(xmmPkt.DFHeader.rawTMTime.rawFine, sizeof(xmmPkt.DFHeader.rawTMTime.rawFine));
      Utils::sunByteSwap(xmmPkt.DFHeader.rawTMTime.rawCoarse, sizeof(xmmPkt.DFHeader.rawTMTime.rawCoarse));

      errstr << "Data Field Time: " 
	     << hex
	     << " Coarse Time " << xmmPkt.DFHeader.TMTime.CoarseTime
	     << " Fine Time " << xmmPkt.DFHeader.TMTime.FineTime
	     << dec
	     << Sas::message(Msg::App, Msg::Noisy);

    }

    // End has time in DFH

  } // End of type of packet (TM, idle or time)

  // SID treatment
  unsigned char rawSID;
  if (bHasTime)
     rawSID=pkt.Data[14];
  else
     rawSID=pkt.Data[8];

  //errstr << "SID: " 
  //  << setw(2) << setfill('0') << (int)rawSID
  //  << Sas::message(Msg::App, Msg::Noisy);

  if (xmmPkt.Header.PacketID.APID != 0) {
     errstr << "Packet ID Info: " 
	 << " APID " << xmmPkt.Header.PacketID.APID
	 << " Type " << xmmPkt.DFHeader.TM.PacketType
	 << " SubType " << xmmPkt.DFHeader.TM.PacketSubtype
         << " SID: " << setw(2) << setfill('0') << (unsigned int)rawSID
         << Sas::message(Msg::App, Msg::Noisy);
  }


  // Specific Processing goes here

  //if (dataProc->acceptPacket(xmmPkt.Header, xmmPkt.DFHeader)) {
  if (dataProc->willacceptPacket(xmmPkt.Header, xmmPkt.DFHeader, rawSID)) {
    dataProc->processPacket(xmmPkt, time);
  }


}

